 #
 #   $Id: TWiki.pm,v 1.4 2005/03/24 23:02:47 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #
 #
 #   Additional Copyrights:
 #
 #    TWiki.pm - Copyright 2004 by David Bright. 
 #		 Based on TWiki.pm from the TWiki distribution:
 #
# Copyright (C) 1999-2003 Peter Thoeny, peter@thoeny.com
#
# Based on parts of Ward Cunninghams original Wiki and JosWiki.
# Copyright (C) 1998 Markus Peter - SPiN GmbH (warpi@spin.de)
# Some changes by Dave Harris (drh@bhresearch.co.uk) incorporated
#
# For licensing info read license.txt file in the TWiki root.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at 
# http://www.gnu.org/copyleft/gpl.html
#
 #	In addition to the compatible licenses, permission to use
 #	TWiki code was also obtained directly from Peter Thoeny:
 #
 #
 #	Date: Tue, 18 May 2004 22:34:34 -0700
 #	From: Peter Thoeny <peter.thoeny@attglobal.net>
 #	To: David A. Bright <BrightD@acm.org>
 #	Cc: TWiki-Core <twiki-core@lists.sourceforge.net>
 #	Subject: Re: OK to use part of TWiki.pm?
 #	                                                                                                    
 #	Dear David,
 #	                                                                                                    
 #	please go ahead and use whatever parts of TWiki you
 #	need. TWiki is an open source project and we never put
 #	any restrictions on using/distributing it (if done
 #	witin the rules of the GPL.) We would appreciate if
 #	you state the originating source with copyright.
 #	                                                                                                    
 #	http://twiki.org/cgi-bin/view/Plugins/PerlDocPlugin
 #	might interest you as well.
 #	                                                                                                    
 #	Good luck with your project, best regards,
 #	Peter
 #

package PodWiki::Formatter::TWiki;

$PodWiki::Formatter::TWiki::VERSION = "1.01";

use lib qw(. lib);

use strict;
#no strict "refs";
use warnings;

use Exporter;

use Data::Dumper;
use Pod::Tree;
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;

use PodWiki::Config;
use PodWiki::Vars;

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS @docpages $config);


@ISA=qw(Exporter PodWiki::Runner PodWiki::Formatter);
@EXPORT=qw();
@EXPORT_OK=qw();
@EXPORT_TAGS=();

#
# Global variables copied out of TWiki.pm (in the TWiki distribution) and then
# modified as little as possible to make it work in PodWiki.
#

# Token character/string that must not occur in any normal text - converted
# to a flag character if it ever does occur (very unlikely).
my $TranslationToken= "\0";	# Null should not be used by any charsets

my $noAutoLink = 0;

my $linkProtocolPattern = "(http|ftp|gopher|news|file|https|telnet)";

# Header patterns based on '+++'. The '###' are reserved for numbered headers
my $headerPatternDa = '^---+(\++|\#+)\s*(.+)\s*$';       # '---++ Header', '---## Header'
my $headerPatternSp = '^\t(\++|\#+)\s*(.+)\s*$';         # '   ++ Header', '   + Header'
my $headerPatternHt = '^<h([1-6])>\s*(.+?)\s*</h[1-6]>'; # '<h6>Header</h6>
my $headerPatternNoTOC = '(\!\!+|%NOTOC%)';  # '---++!! Header' or '---++ Header %NOTOC% ^top'

my	$upperAlpha = "[:upper:]";
my	$lowerAlpha = "[:lower:]";
my	$numeric = "[:digit:]";
my	$mixedAlpha = "[:alpha:]";
my    $mixedAlphaNum = "${mixedAlpha}${numeric}";
my    $lowerAlphaNum = "${lowerAlpha}${numeric}";

my    $wikiWordRegex = qr/[$upperAlpha]+[$lowerAlpha]+[$upperAlpha]+[$mixedAlphaNum]*/;
my    $webNameRegex = qr/[$upperAlpha]+[$lowerAlphaNum]*/;
my    $defaultWebNameRegex = qr/_[${mixedAlphaNum}_]+/;
my    $anchorRegex = qr/\#[${mixedAlphaNum}_]+/;
my    $abbrevRegex = qr/[$upperAlpha]{3,}/;


my $isList;
my @listTypes;
my @listElements;

my $webName = "Main";		# We define a constant "web" name - PodWiki doesn't have separate name spaces
my $topicName = "";

### End of TWiki globals 



sub new {
  my($this, %param) = @_;
  my $class = ref($this) || $this;
  my $self = \%param;
  $config = $self->{config};
  bless($self, $class);
  return $self;
}

sub tags {
  my($this) = @_;
  return qw(twiki);
}

sub do_img {
  my($this, $params) = @_;
  my($link, @params, %params);

  local $_ = $params;

  @params = split; # Split on whitespace to separate keyword=value pairs
  foreach my $param (@params) {
	my($key, @vals) = split(/=/, $param);
	my($val) = join("=", @vals); # Re-assemble values that had embedded equal signs
	$key =~ s/^\s*(\S*)"\s*$/$1/; # Remove any leading and trailing whitespace
	$val =~ s/^\s*"([^"]*)"\s*$/$1/; # Remove any double quotes around value, plus any leading and trailing whitespace
	$params{$key} = $val;
  }

  if (!defined($params{"src"})) {
        $params{"src"} = ""; # Make sure we have a src parameter defined, even if it is empty
  }
  if (!defined($params{"alt"})) {
        $params{"alt"} = $params{"src"}; # Default the alt parameter to the src
  }
         
  @params = ();
  foreach my $key (sort keys(%params)) {
        push(@params, $key . '="' . $params{$key} . '"');
  }
  $link = "<img " . join(",", @params) . " />";

  return $link;
}

sub do_head {
  my($this, $level, $title) = @_;
  my($target) = &PodWiki::Runner::clean_head($title);
  
  if ($title =~ m/^${headerPatternNoTOC}\s*(.*)\s*$/) {
      $title = $2;
     $target = &PodWiki::Runner::clean_head($title);
  } else {
      push @PodWiki::Runner::head, {
	  level => $level,
	  target => $target,
	  title => $title
	  };
  }

  return "<h${level}><a name=\"$target\">$title</a></h${level}>";
}

#
# This is copied out of TWiki.pm (in the TWiki distribution) and then
# modified as little as possible to make it work in PodWiki. TWiki
# plugins are not supported.
#


# =========================
sub takeOutVerbatim
{
    my( $intext, $verbatim ) = @_;
    
    if( $intext !~ /<verbatim>/oi ) {
        return( $intext );
    }
    
    # Exclude text inside verbatim from variable substitution
    
    my $tmp = "";
    my $outtext = "";
    my $nesting = 0;
    my $verbatimCount = $#{$verbatim} + 1;
    
    foreach( split( /\n/, $intext ) ) {
        if( /^(\s*)<verbatim>\s*$/i ) {
            $nesting++;
            if( $nesting == 1 ) {
                $outtext .= "$1%_VERBATIM$verbatimCount%\n";
                $tmp = "";
                next;
            }
        } elsif( m|^\s*</verbatim>\s*$|i ) {
            $nesting--;
            if( ! $nesting ) {
                $verbatim->[$verbatimCount++] = $tmp;
                next;
            }
        }

        if( $nesting ) {
            $tmp .= "$_\n";
        } else {
            $outtext .= "$_\n";
        }
    }
    
    # Deal with unclosed verbatim
    if( $nesting ) {
        $verbatim->[$verbatimCount] = $tmp;
    }
       
    return $outtext;
}

# =========================
# set type=verbatim to get back original text
#     type=pre to convert to HTML readable verbatim text
sub putBackVerbatim
{
    my( $text, $type, @verbatim ) = @_;
    
    for( my $i=0; $i<=$#verbatim; $i++ ) {
        my $val = $verbatim[$i];
        if( $type ne "verbatim" ) {
            $val =~ s/</&lt;/g;
            $val =~ s/</&gt;/g;
            $val =~ s/\t/   /g; # A shame to do this, but been in TWiki.org have converted
                                # 3 spaces to tabs since day 1
        }
        $text =~ s|%_VERBATIM$i%|<$type>\n$val</$type>|;
    }

    return $text;
}

# =========================
sub emitList {
    my( $theType, $theElement, $theDepth ) = @_;
    my $result = "";
    $isList = 1;

    if( @listTypes < $theDepth ) {
        my $firstTime = 1;
        while( @listTypes < $theDepth ) {
            push( @listTypes, $theType );
            push( @listElements, $theElement );
            $result .= "<$theElement>\n" unless( $firstTime );
            $result .= "<$theType>\n";
            $firstTime = 0;
        }

    } elsif( @listTypes > $theDepth ) {
        while( @listTypes > $theDepth ) {
            local($_) = pop @listElements;
            $result .= "</$_>\n";
            local($_) = pop @listTypes;
            $result .= "</$_>\n";
        }
        $result .= "</$listElements[$#listElements]>\n" if( @listElements );

    } elsif( @listElements ) {
        $result = "</$listElements[$#listElements]>\n";
    }

    if( ( @listTypes ) && ( $listTypes[$#listTypes] ne $theType ) ) {
        $result .= "</$listTypes[$#listTypes]>\n<$theType>\n";
        $listTypes[$#listTypes] = $theType;
        $listElements[$#listElements] = $theElement;
    }

    return $result;
}

# =========================
sub emitTR {
    my ( $thePre, $theRow, $insideTABLE ) = @_;

    my $text = "";
    my $attr = "";
    my $l1 = 0;
    my $l2 = 0;
    if( $insideTABLE ) {
        $text = "$thePre<tr>";
    } else {
        $text = "$thePre<table border=\"1\" cellspacing=\"0\" cellpadding=\"1\"> <tr>";
    }
    $theRow =~ s/\t/   /g;  # change tabs to space
    $theRow =~ s/\s*$//;    # remove trailing spaces
    $theRow =~ s/(\|\|+)/$TranslationToken . length($1) . "\|"/ge;  # calc COLSPAN
    foreach( split( /\|/, $theRow ) ) {
        $attr = "";
        #AS 25-5-01 Fix to avoid matching also single columns
        if ( s/$TranslationToken([0-9]+)// ) { # No o flag for mod-perl compatibility
            $attr = " colspan=\"$1\"" ;
        }
        s/^\s+$/ &nbsp; /;
        /^(\s*).*?(\s*)$/;
        $l1 = length( $1 || "" );
        $l2 = length( $2 || "" );
        if( $l1 >= 2 ) {
            if( $l2 <= 1 ) {
                $attr .= ' align="right"';
            } else {
                $attr .= ' align="center"';
            }
        }
        if( /^\s*(\*.*\*)\s*$/ ) {
            $text .= "<th$attr> $1 </th>";
        } else {
            $text .= "<td$attr> $_ </td>";
        }
    }
    $text .= "</tr>";
    return $text;
}

# =========================
sub getRenderedVersion {
    my( $this, $text, $theWeb, $meta ) = @_;
    my( $head, $result, $extraLines, $insidePRE, $insideTABLE, $insideNoAutoLink );

    return "" unless $text;  # nothing to do

    # FIXME: Get $theTopic from parameter to handle [[#anchor]] correctly
    # (fails in %INCLUDE%, %SEARCH%)
    my $theTopic = $topicName;

    # PTh 22 Jul 2000: added $theWeb for correct handling of %INCLUDE%, %SEARCH%
    if( !$theWeb ) {
        $theWeb = $webName;
    }

    $head = "";
    $result = "";
    $insidePRE = 0;
    $insideTABLE = 0;
    $insideNoAutoLink = 0;      # PTh 02 Feb 2001: Added Codev.DisableWikiWordLinks
    $isList = 0;
    @listTypes = ();
    @listElements = ();

    # Convert leading spaces to tabs. In TWiki, this would be done during the
    # save code. For PodWiki, it needs to be done here.
    $text =~ s/^(( {3})+)/"\t" x (length($1)\/3)/geom; 

    # Initial cleanup
    $text =~ s/\r//g;
    $text =~ s/(\n?)$/\n<nop>\n/s; # clutch to enforce correct rendering at end of doc
    $text =~ s/$TranslationToken/!/go;	# Convert any occurrences of token
    					# (very unlikely)

    my @verbatim = ();
    $text = takeOutVerbatim( $text, \@verbatim );
    $text =~ s/\\\n//g;  # Join lines ending in "\"

    # do not render HTML head, style sheets and scripts
    if( $text =~ m/<body[\s\>]/i ) {
        my $bodyTag = "";
        my $bodyText = "";
        ( $head, $bodyTag, $bodyText ) = split( /(<body)/i, $text, 3 );
        $text = $bodyTag . $bodyText;
    }
    
    # Wiki Plugin Hook --- Not supported in PodWiki
    #&TWiki::Plugins::startRenderingHandler( $text, $theWeb, $meta );

    foreach( split( /\n/, $text ) ) {

        # change state:
        m|<pre>|i  && ( $insidePRE = 1 );
        m|</pre>|i && ( $insidePRE = 0 );
        m|<noautolink>|i   && ( $insideNoAutoLink = 1 );
        m|</noautolink>|i  && ( $insideNoAutoLink = 0 );

        if( $insidePRE ) {
            # inside <PRE>

            # close list tags if any
            if( @listTypes ) {
                $result .= &emitList( "", "", 0 );
                $isList = 0;
            }

# Wiki Plugin Hook --- Not supported in PodWiki
            #&TWiki::Plugins::insidePREHandler( $_ );

            s/(.*)/$1\n/;
            s/\t/   /g;		# Three spaces
            $result .= $_;

        } else {
          # normal state, do Wiki rendering

# Wiki Plugin Hook --- Not supported in PodWiki
	  #&TWiki::Plugins::outsidePREHandler( $_ );
          $extraLines = undef;   # Plugins might introduce extra lines
          do {                   # Loop over extra lines added by plugins
            $_ = $extraLines if( defined $extraLines );
            s/^(.*?)\n(.*)$/$1/s;
            $extraLines = $2;    # Save extra lines, need to parse each separately

# Blockquote
            s/^>(.*?)$/> <cite> $1 <\/cite><br \/>/g;

# Embedded HTML
            s/\<(\!\-\-)/$TranslationToken$1/g;  # Allow standalone "<!--"
            s/(\-\-)\>/$1$TranslationToken/g;    # Allow standalone "-->"
            s/(\<\<+)/"&lt\;" x length($1)/ge;
            s/(\>\>+)/"&gt\;" x length($1)/ge;
            s/\<nop\>/nopTOKEN/g;  # defuse <nop> inside HTML tags
            s/\<(\S.*?)\>/$TranslationToken$1$TranslationToken/g;
            s/</&lt\;/g;
            s/>/&gt\;/g;
            s/$TranslationToken(\S.*?)$TranslationToken/\<$1\>/go;
            s/nopTOKEN/\<nop\>/g;
            s/(\-\-)$TranslationToken/$1\>/go;
            s/$TranslationToken(\!\-\-)/\<$1/go;

# Handle embedded URLs
            s!(^|[\-\*\s\(])($linkProtocolPattern\:([^\s\<\>\"]+[^\s\.\,\!\?\;\:\)\<]))!$1 . $this->pod_link($2)!geo;

# Entities
            s/&(\w+?)\;/$TranslationToken$1\;/g;      # "&abc;"
            s/&(\#[0-9]+)\;/$TranslationToken$1\;/g;  # "&#123;"
            s/&/&amp;/g;                              # escape standalone "&"
            s/$TranslationToken/&/go;

# Headings
            # '<h6>...</h6>' HTML rule
            s/$headerPatternHt/$this->do_head($1, $2)/geoi;
            # '\t+++++++' rule
            s/$headerPatternSp/$this->do_head((length($1)), $2)/geo;
            # '----+++++++' rule
            s/$headerPatternDa/$this->do_head((length($1)), $2)/geo;

# Horizontal rule
            s/^---+/<hr \/>/;
            s!^([a-zA-Z0-9]+)----*!<table width=\"100%\"><tr><td valign=\"bottom\"><h2>$1</h2></td><td width=\"98%\" valign=\"middle\"><hr /></td></tr></table>!o;

# Table of format: | cell | cell |
            # PTh 25 Jan 2001: Forgiving syntax, allow trailing white space
            if( $_ =~ /^(\s*)\|.*\|\s*$/ ) {
                s/^(\s*)\|(.*)/&emitTR($1,$2,$insideTABLE)/e;
                $insideTABLE = 1;
            } elsif( $insideTABLE ) {
                $result .= "</table>\n";
                $insideTABLE = 0;
            }

# Lists and paragraphs
            s/^\s*$/\n\n/o                 && ( $isList = 0 ); # In TWiki, inserts <p />; PodWiki uses the \n\n (see format())
            m/^(\S+?)/o                      && ( $isList = 0 );
            s/^(\t+)(\S+?):\s/<dt> $2<\/dt><dd> /o && ( $result .= &emitList( "dl", "dd", length $1 ) );
            s/^(\t+)\* /<li> /o              && ( $result .= &emitList( "ul", "li", length $1 ) );
            s/^(\t+)\d+\.? ?/<li> /o         && ( $result .= &emitList( "ol", "li", length $1 ) );
            if( ! $isList ) {
                $result .= &emitList( "", "", 0 );
                $isList = 0;
            }

# '#WikiName' anchors
            s/^(\#)($wikiWordRegex)/ '<a name="' . $this->clean_head( $2 ) . '"><\/a>'/ge;

# enclose in white space for the regex that follow
             s/(.*)/\n$1\n/;

# Emphasizing
            # PTh 25 Sep 2000: More relaxed rules, allow leading '(' and trailing ',.;:!?)'
            s/([\s\(])==([^\s]+?|[^\s].*?[^\s])==([\s\,\.\;\:\!\?\)])/$1 . "<tt>" . $2 . "<\/tt>" . $3/ge;
            s/([\s\(])__([^\s]+?|[^\s].*?[^\s])__([\s\,\.\;\:\!\?\)])/$1<strong><em>$2<\/em><\/strong>$3/g;
            s/([\s\(])\*([^\s]+?|[^\s].*?[^\s])\*([\s\,\.\;\:\!\?\)])/$1<strong>$2<\/strong>$3/g;
            s/([\s\(])_([^\s]+?|[^\s].*?[^\s])_([\s\,\.\;\:\!\?\)])/$1<em>$2<\/em>$3/g;
            s/([\s\(])=([^\s]+?|[^\s].*?[^\s])=([\s\,\.\;\:\!\?\)])/$1 . "<tt>" . $2 . "<\/tt>" . $3/ge;

# Mailto
	    # Email addresses must always be 7-bit, even within I18N sites

	    # RD 27 Mar 02: Mailto improvements - FIXME: check security...
	    # Explicit [[mailto:... ]] link without an '@' - hence no 
	    # anti-spam padding needed.
            # '[[mailto:string display text]]' link (no '@' in 'string'):
            s/\[\[mailto\:([^\s\@]+)\s+(.+?)\]\]/$this->pod_link("$2|mailto:$1")/ge;

	    # Explicit [[mailto:... ]] link including '@', with anti-spam 
	    # padding, so match name@subdom.dom.
            # '[[mailto:string display text]]' link
            s/\[\[mailto\:([a-zA-Z0-9\-\_\.\+]+)\@([a-zA-Z0-9\-\_\.]+)\.(.+?)(\s+|\]\[)(.*?)\]\]/$this->pod_link("$5|mailto:$1\@$2.$3")/ge;

	    # Normal mailto:foo@example.com ('mailto:' part optional)
	    # FIXME: Should be '?' after the 'mailto:'...
            s/([\s\(])(?:mailto\:)*([a-zA-Z0-9\-\_\.\+]+)\@([a-zA-Z0-9\-\_\.]+)\.([a-zA-Z0-9\-\_]+)(?=[\s\.\,\;\:\!\?\)])/$1 . $this->pod_link("mailto:$2\@$3.$4")/ge;

# Make internal links
	    # Spaced-out Wiki words with alternative link text
            # '[[Web.odd wiki word#anchor][display text]]' link:
            s/\[\[(\#[^\]]+)\]\[([^\]]+)\]\]/$this->pod_link("$2|" . $PodWiki::Runner::page . "$1")/ge; # Internal page anchor - need to prefix with page name
            s/\[\[([^\]]+)\]\[([^\]]+)\]\]/$this->pod_link("$2|$1")/ge;
            # RD 25 Mar 02: Codev.EasierExternalLinking
            # '[[URL#anchor display text]]' link:
            s/\[\[([a-z]+\:\S+)\s+(.*?)\]\]/$this->pod_link("$2|$1")/ge;
	    # Spaced-out Wiki words
            # '[[Web.odd wiki word#anchor]]' link:
            s/\[\[([^\]]+)\]\]/$this->pod_link($1)/ge;

            # do normal WikiWord link if not disabled by <noautolink> or NOAUTOLINK preferences variable
            unless( $noAutoLink || $insideNoAutoLink ) {

                # 'Web.TopicName#anchor' link:
                s/([\s\(])($webNameRegex)\.($wikiWordRegex)($anchorRegex)/$1 . $this->pod_link($2 . "." . $3 . "#" . $4)/geo;
                # 'Web.TopicName' link:
                s/([\s\(])($webNameRegex)\.($wikiWordRegex)/$1 . $this->pod_link($2 . "." . $3)/geo;

                # 'TopicName#anchor' link:
                s/([\s\(])($wikiWordRegex)($anchorRegex)/$1 . $this->pod_link( $2 . "#" . $3)/geo; # If webs were supported: $1 . $this->pod_link($theWeb . "." . $2 . "#" . $3)

                # 'TopicName' link:
		s/([\s\(])($wikiWordRegex)/$1 . $this->pod_link($2)/geo; # If webs were supported: $1 . $this->pod_link($theWeb . "." . $2)

		# PodWiki Note - ABBREV links are disabled. They can be re-enabled by uncommenting the two substitution commands below.
		# Handle acronyms/abbreviations of three or more letters
                # 'Web.ABBREV' link:
                #s/([\s\(])($webNameRegex)\.($abbrevRegex)/$1 . $this->pod_link($2 . "." . $3)/geo;
                # 'ABBREV' link:
		#s/([\s\(])($abbrevRegex)/$1 . $this->pod_link($2)/geo; # If webs were supported: $1 . $this->pod_link($theWeb . "." . $2)

                s/$TranslationToken(\S.*?)$TranslationToken/$1/go;
            }

            s/^\n//;
            s/\t/   /g;
            $result .= $_;

          } while( defined( $extraLines ) );  # extra lines produced by plugins
        }
    }
    if( $insideTABLE ) {
        $result .= "</table>\n";
    }
    $result .= &emitList( "", "", 0 );
    if( $insidePRE ) {
        $result .= "</pre>\n";
    }

    # Wiki Plugin Hook --- Not supported in PodWiki
    #&TWiki::Plugins::endRenderingHandler( $result );

    $result = putBackVerbatim( $result, "pre", @verbatim );

    $result =~ s|\n?<nop>\n$||o; # clean up clutch


    #
    # Variable substitution
    #
    $result =~ s/\$/$TranslationToken/g;	# Save anything that looks like a PodWiki variable substitution.
    $result =~ s/%BR%/<br \/>/g;
    $result =~ s/%TOC(\s*{.*})?\s*%/\n\nP<intern:toc>\n\n/g;
    $result =~ s/%WEB%/$theWeb/g;
    $result =~ s/%ATTACHURL%/$config->{gfx}/g;
    $result =~ s/%WIKITOOLNAME%/PodWiki/g;
    $result =~ s/%TOPIC%/\$page/g;
    $result =~ s/%STARTINCLUDE%//g;
    $result =~ s/%STOPINCLUDE%//g;
    # Doc graphics; when the "standard" TWikiPreferences file defines one that is functionally equivalent to
    # a PodWiki one, we substitute. Otherwise, we just let the general substitution rule (below) take effect.
    $result =~ s/%H%/\$help/g;
    $result =~ s/%I%/\$idea/g;
    #$result =~ s/%M%/\$rtarrow/g;
    $result =~ s/%N%/\$RED<font size="80%"><strong>NEW<\/strong><\/font>\$END/g;
    #$result =~ s/%P%/\$pencil/g;
    $result =~ s/%Q%/\$help/g;
    $result =~ s/%S%/\$bookmark/g;
    $result =~ s/%T%/\$idea/g;
    $result =~ s/%U%/\$ORANGE<font size="80%"><strong>UPDATED<\/strong><\/font>\$END/g;
    #$result =~ s/%X%/\$warning/g;
    $result =~ s/%Y%/\$ok/g;
    # Color variables
    $result =~ s/%ENDCOLOR%/\$END/g;
    $result =~ s/%YELLOW%/\$YELLOW/g;
    $result =~ s/%ORANGE%/\$ORANGE/g;
    $result =~ s/%RED%/\$RED/g;
    $result =~ s/%PINK%/\$PINK/g;
    $result =~ s/%PURPLE%/\$PURPLE/g;
    $result =~ s/%TEAL%/\$TEAL/g;
    $result =~ s/%NAVY%/\$NAVY/g;
    $result =~ s/%BLUE%/\$BLUE/g;
    $result =~ s/%AQUA%/\$AQUA/g;
    $result =~ s/%LIME%/\$LIME/g;
    $result =~ s/%GREEN%/\$GREEN/g;
    $result =~ s/%OLIVE%/\$OLIVE/g;
    $result =~ s/%MAROON%/\$MAROON/g;
    $result =~ s/%BLACK%/\$BLACK/g;
    $result =~ s/%GRAY%/\$GRAY/g;
    $result =~ s/%SILVER%/\$SILVER/g;
    $result =~ s/%WHITE%/\$WHITE/g;

    # include tag, prepare for split (see below)
    # We have to throw away the start/stop markers because that is not implemented by POD
    $result =~ s/%INCLUDE\s*\{\s*\"([^\"]*)\"([^\}]*)\s*\}%/\n\nP<$1>\n\n/g;

    $result =~ s/%([A-Z]*)%/\$$1/g;

    $result = PodWiki::Runner::interpolate($result);

    # Convert graphics resulting from variable interpolation to HTML
    $result =~ s/G<([^>]*)>/<img src="$config->{gfx}\/$1" alt="$1" \/>/g;

    $result =~ s/$TranslationToken/\$/g;	# Restore things that looked like PodWiki variables


    return "$head$result";
}

sub format {
  my($this, $text) = @_;
  local $_ = $text;

  s/%META:.*?%//g; # Delete META tags

  $_ = $this->getRenderedVersion($_, $webName, "");

  # make paragraphs of the text
  # and do the post-processing
  my @nodes = split /\n{2,}/;
  my $out = "";
  my $islist = 0;
  my @list = ();
  foreach my $node (@nodes) {
    next if($node =~ /^\s*$/s);
    if ($node =~ /^<h\d>/) {
      # head, feed out unchanged
      print "$node\n";
    }
#    elsif ($node =~ /^\s/m) {
#      # verbatim
#      print "<pre>$node</pre>\n";
#    }
    elsif ($node =~ /^(P<[^>]*>)$/s) {
      # include another page
      my $tree = new Pod::Tree;
      $tree->load_string("P<$1>");
      $tree->walk(\&PodWiki::Runner::walker);
    }
    else {
      print "<p>$node</p>\n";
    }
  }
}

END {
}

1; #end

__END__

=head1 NAME

PodWiki::Formatter::TWiki - PodWiki formatter that understands (much of) the TWiki markup language

=head1 SYNOPSIS

  use PodWiki::Formatter::TWiki;
  
  $fmtr = new Pod::Formatter::TWiki;
  @tags = $fmtr->tags;
  $fmtr->format($text);
  

=head1 REQUIRES

Perl 5.6.0, Exporter, IO::File, Pod::Tree::Node


=head1 EXPORTS

Nothing


=head1 DESCRIPTION

C<PodWiki::Formatter::TWiki> parses text that uses the TWiki markup
language and prints the equivalent HTML. This is intended to be used
by PodWiki users that are migrating content from a TWiki
system. See L<http://www.TWiki.org/> for a description of the markup
language. Not all features of TWiki are supported, including (but
not limited to):

=over 4

=item Most plugins, which take the form of: %NAME(args)%

=back


=head1 METHODS

=over 4

=item I<$fmtr> = C<new> C<PodWiki::Formatter::TWiki>

Creates a new C<PodWiki::Formatter::TWiki> object.

=item I<@tags> = I<$fmtr>->C<tags>()

Returns an array containing the tags that are processed by the formatter.

=item I<$ok> = I<$fmtr>->C<format>(I<$text>)

Parses I<$text> according to the rules of the formatter. The resultant
text, if any, is written to STDOUT.

=back

=head1 DIAGNOSTICS

=over 4

=back


=head1 NOTES

None.

=head1 SEE ALSO

perl(1), L<C<PodWiki::Runner>>


=head1 ACKNOWLEDGMENTS

=over 4

=item *

Peter Thoeny, peter@thoeny.com

=item *

The TWiki Community L<http://www.twiki.org>

=back


=head1 AUTHOR

David Bright <BrightD@acm.org>


=head1 COPYRIGHT

Copyright 2004 by David Bright. This module is free software; you can
redistribute it and/or modify it under the same terms as Perl itself.

Based on TWiki.pm from the TWiki distribution:

Copyright (C) 1999-2003 Peter Thoeny, peter@thoeny.com

Based on parts of Ward Cunninghams original Wiki and JosWiki.
Copyright (C) 1998 Markus Peter - SPiN GmbH (warpi@spin.de)
Some changes by Dave Harris (drh@bhresearch.co.uk) incorporated

For licensing info read license.txt file in the TWiki root.
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details, published at 
http://www.gnu.org/copyleft/gpl.html

