 #
 #   $Id: Tiki.pm,v 1.17 2005/03/24 23:02:47 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #
 #
 #   Additional Copyrights:
 #
 #    Tiki.pm - Copyright 2004 by David Bright
 #

package PodWiki::Formatter::Tiki;

$PodWiki::Formatter::Tiki::VERSION = "2.03";

use lib qw(. lib);

use strict;
#no strict "refs";
use warnings;

use Exporter;

use Data::Dumper;
use Pod::Tree;
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;

use PodWiki::Config;
use PodWiki::Vars;

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS @docpages $config);


@ISA=qw(Exporter PodWiki::Runner PodWiki::Formatter);
@EXPORT=qw();
@EXPORT_OK=qw();
@EXPORT_TAGS=();

sub new {
  my($this, %param) = @_;
  my $class = ref($this) || $this;
  my $self = \%param;
  $config = $self->{config};
  bless($self, $class);
  return $self;
}

sub tags {
  my($this) = @_;
  return qw(tiki);
}

sub do_img {
  my($this, $params) = @_;
  my($link, @params, %params);

  local $_ = $params;

  @params = split; # Split on whitespace to separate keyword=value pairs
  foreach my $param (@params) {
	my($key, @vals) = split(/=/, $param);
	my($val) = join("=", @vals); # Re-assemble values that had embedded equal signs
	$key =~ s/^\s*(\S*)"\s*$/$1/; # Remove any leading and trailing whitespace
	$val =~ s/^\s*"([^"]*)"\s*$/$1/; # Remove any double quotes around value, plus any leading and trailing whitespace
	$params{$key} = $val;
  }

  if (!defined($params{"src"})) {
	$params{"src"} = ""; # Make sure we have a src parameter defined, even if it is empty
  }
  if (!defined($params{"alt"})) {
	$params{"alt"} = $params{"src"}; # Default the alt parameter to the src
  }
	
  @params = ();
  foreach my $key (sort keys(%params)) {
	push(@params, $key . '="' . $params{$key} . '"');
  } 
  $link = "<img " . join(",", @params) . " />";
  return $link;
}

sub do_head {
  my($this, $level, $title) = @_;
  my($target) = &PodWiki::Runner::clean_head($title);
  
  push @PodWiki::Runner::head, {
      level => $level,
      target => $target,
      title => $title
      };

  return "<h${level}><a name=\"$target\">$title</a></h${level}>";
}

sub format {
  my($this, $text) = @_;
  local $_ = $text;

  $_ ="\n\n$_\n";

  # Scarf out preformatted sections and replace with a unique key
  my @pp = ();
  s/\~pp\~((.|\n)*?)\~\/pp\~/"~pp~" . (push(@pp, $1), $#pp) . "~\/pp~"/ge;
  # Scarf out non-parsed sections and replace with a unique key
  my @np = ();
  s/\~np\~((.|\n)*?)\~\/np\~/"~np~" . (push(@np, $1), $#np) . "~\/np~"/ge;


  # heads:
  # !Title
  my %h = ( '!' => 1, '!!' => 2, '!!!' => 3, '!!!!' => 4, '!!!!!' => 5, '!!!!!!' => 6);
#  s/(?<=\n)(!)+[-+]?(.*)\s*\n/\n\n<h$h{$1}>$2<\/h$h{$1}>\n\n/g;
  s/(?<=\n)(!+)[-+]?\s*(.*)\s*\n/"\n\n" . $this->do_head($h{$1}, $2) . "\n\n"/ge;

  # table of contents
  s/{maketoc}/\n\nP<intern:toc>\n\n/g;

  # formats:
  # __bold__
  # ''italic''
  s/(?<!\\)__(.*?)__/<b>$1<\/b>/g;           # bold
  s/(?<![<\/])''(.*?)(?<!<)''/<i>$1<\/i>/g;  # italic

  # image var post-processing
  # $ok  (maintained by PodWiki::Vars) => {picture file=ok.png}
  s/G<([^>]*)>/{picture file=$1}/g;

  # images:
  # {picture ...}
  # {img ...}
  s/\{picture file=([^\}]+)\}/<img src="$config->{gfx}\/$1" alt="$1" \/>/g;
  s/({img\s+([^\}]+)})/$this->do_img($2)/ge;

  # Replace boxes
  s/\^([^\^]+)\^/<div class="simplebox">$1<\/div>/g;
  # Replace colors ~~color:text~~
  s/\~\~([^\:]+):([^\~]+)\~\~/<span style="color:$1;">$2<\/span/g;
  # Underlined text
  s/===([^\=]+)===/<span style="text-decoration:underline;">$1<\/span>/g;
  # Center text
  s/::([^\:]+)::/<div align="center">$1<\/div>/g;
  # Monospace text
  s/-\+(.*?)\+-/<tt>$1<\/tt>/g;

  # Titlebars
  s/(^|\n)-=(.+?)=-/$1<div class='titlebar'>$2<\/div>\n/g;
  # Replace definition lists
  s/(^|\n);([^:]+):([^\n]+)/$1<dl><dt>$2<\/dt><dd>$3<\/dd><\/dl>/g;


  my $page_regex = '[A-Za-z0-9_\x80-\xFF][\.: A-Za-z0-9_\-\x80-\xFF]*[A-Za-z0-9_\x80-\xFF]';
  # CamelCase links
  #page_regex = '[A-Z][a-z0-9_\-]+[A-Z][a-z0-9_\-]+[A-Za-z0-9\-_]*';
  s/([ \n\t\r\,\;]|^)([A-Z][a-z0-9_\-]+[A-Z][a-z0-9_\-]+[A-Za-z0-9\-_]*)($|[ \n\t\r\,\;\.])/$1 . $this->pod_link($2) . $3/ge;
  # Now, get rid of the )) (( around CamelCase non-links
  s/([ \n\t\r\,\;>]|^)\)\)([^\(]+)\(\(($|[ \n\t\r\,\;\.<])/$1$2$3/g;

  # Free links - reverse the order of the target and description
  s/\(\(($page_regex)\|([^\)]+)\)\)/$this->pod_link($2 . "|" . $1)/ge;
  # Link with no description
  s/\(\(($page_regex)\)\)/$this->pod_link($1)/ge;

  # External links - reverse the order of the target and description
  s/([^\[])\[([^\[\]|]+)\|([^\]]*)\]/$1 . $this->pod_link($3 . "|" . $2)/ge;
  # Link with no description
  s/([^\[])\[([^\[\]]+)\]/$1 . $this->pod_link($2)/ge;
  # Bracketed text
  s/\[\[([^\]]*)\]/[$1]/gi;

  # include tag, prepare for split (see below)
  # We have to throw away the start/stop markers ($3, $5 in regexp below) because that is not implemented by POD
  s/\{INCLUDE\s*\(\s*page\s*=>\s*\"?([^,\)\"]*)\"?\s*(,\s*start\s*=>([^,\)]*)){0,1}(,\s*stop\s*=>(^[^,\)]*)){0,1}\)\}\s*\{INCLUDE\}/\n\nP<$1>\n\n/g;

  # Horizontal lines
  s/(^|\n)---.*?($|\n)/$1<hr \/>$2/g;
		
  # resolve entities
  s/E<([^>]*)>/&$1;/g;


  # Tables
  {
      my $intable = 0;
      my $newdata;
      my $col_count;
      my $row_number;

      foreach my $line (split /\n/) {
	  my $table_end_found = 0;
	  if (! $intable ) {
	      if ($line =~ m/^\|\|(.*)$/) {
		  # Beginning of table, strip the table beginning mark and set in-table variables
		  $line = $1;
		  $col_count = 0;
		  $row_number = 0;
		  $intable = 1;
		  $newdata .= "\n\n<table>\n";
	      } else {
		  $newdata .= $line . "\n";	# Didn't begin table, so just add line to data
	      }
	  }
	  if ($intable) {
	      my $cols_done;
	      if ($line =~ m/^(.*)\|\|\s*$/) {
		  # This line ends the table; strip the table ending mark
		  $line = $1;
		  $table_end_found = 1;
	      }
	      if (($line !~ m/^\s*$/)) {
		  # Process non-empty lines
		  my @cols;
		  $row_number++;
		  if (0 == ($row_number % 2)) {
		      $newdata .= '<tr class="even">';
		  } else {
		      $newdata .= '<tr class="odd">';
		  }
		  @cols = split(/\|/, $line); # Split the line into columns
		  if ($col_count < ($#cols + 1)) { # Get the highest column count (for column spanning)
		      $col_count = $#cols + 1;
		  }
		  $cols_done = 0;
		  foreach my $col (@cols) {
		      my $span = "";
		      if ($cols_done == $#cols) {
			  # This is the last column, check for column spanning
			  if ($cols_done + 1 < $col_count) {
			      $span = " colspan=\"" . ($col_count - $cols_done) . "\"";
			  }
		      }
		      $newdata .= "<td$span>" . $col . "</td>";
		      $cols_done++;
		  }
	      }
	      $newdata .= "</tr>\n";
	      if ($table_end_found) {
		  $newdata .= "</table>\n\n";
		  $intable = 0;
	      }
	  }
      }
      $_ = $newdata;
  }

  # Lists
  {
      my $curlistlevel = 0;
      my $newlistlevel = 0;
      my @closelist = ();
      my $newdata;
      my $listart;
      my $liend;
      my $liitem;
  
    foreach my $node (split /\n/) {
	  my $line = $node;
	  my $litype;

	  $listart = "";	# Stuff to put at the beginning of the line (blank or <li>)
	  $liend = "";		# Stuff to put at the end of the line (blank or </li>)
	  $newlistlevel = 0;
	  if ($line =~ m/^((\*|\#)+)(.*)$/) {
	      $litype = $1;
	      $liitem = $3;
	      $newlistlevel = length($litype);
	  }
	  
	  if ($newlistlevel > 0) {
	      my $listopen;
	      my $listclose;
	      
	      $listart = '<li>';
	      $liend = '</li>';
	      $line = $liitem;
	      if (substr($litype, 0, 1) eq '*') {
		  $listopen = "<ul>";
		  $listclose = "</ul>";
	      } elsif (substr($litype, 0, 1) eq '#') {
		  $listopen = "<ol>";
		  $listclose = "</ol>";
	      } else {
		  $newdata .= "\n\nUnknown list type " . substr($litype, 0, 1) . " - please file a bug report.\n\n";
		  $listopen = "<ul>"; # Use the ul list type
		  $listclose = "</ul>";
	      }
	      if ($newlistlevel < $curlistlevel) {
		  while ($newlistlevel < $curlistlevel) {
		      $newdata .= shift(@closelist) . "\n";
		      $curlistlevel--;
		  }
	      }
	      if ($newlistlevel > $curlistlevel) {
		  while ($newlistlevel > $curlistlevel) {
		      $newdata .= $listopen . "\n";
		      unshift(@closelist, $listclose);
		      $curlistlevel++;
		  }
	      }
	  } elsif ($curlistlevel > 0) {
	      # We are no longer in a list; close out the existing lists
	      while ($curlistlevel > 0) {
		  $newdata .= shift(@closelist) . "\n";
		  $curlistlevel--;
	      }
	  }
	  $newdata .= $listart . $line . $liend . "\n";
      }
      if ($curlistlevel > 0) {
	  # There are open lists; close them
	  while ($curlistlevel > 0) {
	      $newdata .= shift(@closelist) . "\n";
	      $curlistlevel--;
	  }
      }
      $_ = $newdata;
  }

  # Replace preformatted sections
  s/\~pp\~(\d+?)\~\/pp\~/"<pre>" . $pp[$1] . "<\/pre>"/ge;
  # Replace non-parsed sections
  s/\~np\~(\d+?)\~\/np\~/$np[$1]/ge;

  # make paragraphs of the text
  # and do the post-processing
  my @nodes = split /\n{2,}/;
  my $out = "";
  my $islist = 0;
  my @list = ();
  foreach my $node (@nodes) {
    next if($node =~ /^\s*$/s);
    if ($node =~ /^<h\d>/) {
      # head, feed out unchanged
      print "$node\n";
    }
    elsif ($node =~ /^\s/m) {
      # verbatim
      print "<pre>$node</pre>\n";
    }
    elsif ($node =~ /^(P<[^>]*>)$/s) {
      # include another page
      my $tree = new Pod::Tree;
      $tree->load_string("P<$1>");
      $tree->walk(\&PodWiki::Runner::walker);
    }
    else {
      print "<p>$node</p>\n";
    }
  }
}

END {
}

1; #end

__END__

=head1 NAME

PodWiki::Formatter::Tiki - PodWiki formatter that understands (much of) the TikiWiki markup language

=head1 SYNOPSIS

  use PodWiki::Formatter::Tiki;
  
  $fmtr = new Pod::Formatter::Tiki;
  @tags = $fmtr->tags;
  $fmtr->format($text);
  

=head1 REQUIRES

Perl 5.6.0, Exporter, IO::File, Pod::Tree::Node


=head1 EXPORTS

Nothing


=head1 DESCRIPTION

C<PodWiki::Formatter::Tiki> parses text that uses the TikiWiki markup
language and prints the equivalent HTML. This is intended to be used
by PodWiki users that are migrating content from a TikiWiki
system. See L<http://tikiwiki.org/> for a description of the markup
language. Not all features of TikiWiki are supported, including (but
not limited to):

=over 4

=item Most plugins, which take the form of: {NAME(args)}...{NAME}

=item rss feeds

=item dynamic content

=back


=head1 METHODS

=over 4

=item I<$fmtr> = C<new> C<PodWiki::Formatter::Tiki>

Creates a new C<PodWiki::Formatter::Tiki> object.

=item I<@tags> = I<$fmtr>->C<tags>()

Returns an array containing the tags that are processed by the formatter.

=item I<$ok> = I<$fmtr>->C<format>(I<$text>)

Parses I<$text> according to the rules of the formatter. The resultant
text, if any, is written to STDOUT.

=back

=head1 DIAGNOSTICS

=over 4

=back


=head1 NOTES

None.

=head1 SEE ALSO

perl(1), L<C<PodWiki::Runner>>


=head1 ACKNOWLEDGMENTS

=over 4

=item *

The TikiWiki Community L<http://tikiwiki.org>

=back


=head1 AUTHOR

David Bright <BrightD@acm.org>


=head1 COPYRIGHT

Copyright 2004 by David Bright. This module is free software; you can
redistribute it and/or modify it under the same terms as Perl itself.
