 #
 #   $Id: WikiShorthand.pm,v 1.8 2005/03/24 23:02:47 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #
 #   Additional Copyrights:
 #
 #    Copyright 2004 by David Bright
 #


package PodWiki::Formatter::WikiShorthand;

$PodWiki::Formatter::WikiShorthand::VERSION = "1.01";

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS @docpages);
use vars qw($me $config $page);

@ISA=qw(Exporter PodWiki::Runner PodWiki::Formatter);
@EXPORT=qw();
@EXPORT_OK=qw();
@EXPORT_TAGS=();

sub new {
  my($this, %param) = @_;
  my $class = ref($this) || $this;
  my $self = \%param;
  $config = $self->{config};
  bless($self, $class);
  return $self;
}

sub tags {
  my($this) = @_;
  return qw(wikishorthand natural human);
}

sub format {
  my ($this, $text) = @_;
  local $_ = $text;
  $_ ="\n\n$_\n";

  # heads:
  # Title
  # -----
  my %h = ( '*' => 1, '=' => 2, '-' => 3, '~' => 4);
  s/(?<=\n)([^\n]+)\n(\-|\*|=|~){3,}/<h$h{$2}>$1<\/h$h{$2}>\n/g;

  # formats:
  # *bold*
  # /italic/
  s/(?<!\\)\*([^\*]+)\*/<b>$1<\/b>/g;           # bold
  s/(?<![<\/])\/([^\/]+)(?<!<)\//<i>$1<\/i>/g;  # italic

  # image var post-processing
  # $ok  (maintained by PodWiki::Vars) => G<ok.png>
  s/G<([^>]*)>/$1/g;

  # images:
  # image.png
  s/([\w\-]*\.png)/<img src="$config->{gfx}\/$1" alt="$1" \/>/g;

  # PodWiki links
  s/\[([^\]]*)\]/$this->pod_link($1)/ge;

  # HTTP links
  s/((http|https|ftp)(:\/\/[^\s]*))/$this->pod_link($1)/gie;
  s/(mailto:[^\s]*)/$this->pod_link($1)/gie;

  # include tag, prepare for split (see below)
  s/(P<[^>]*>)/\n\n$1\n\n/g;

  # resolve entities
  s/E<([^>]*)>/&$1;/g;

  # make paragraphs of the text
  # and do the post-processing
  my @nodes = split /\n{2,}/;
  my $out = "";
  my $islist = 0;
  my @list = ();
  foreach my $node (@nodes) {
    next if($node =~ /^\s*$/s);
    if ($node =~ /^<h\d>/) {
      # head, feed out unchanged
      print "$node\n";
    }
    elsif ($node =~ /^o /s) {
      # list item
      $node =~ s/^o //s;
      push @list, $node;
      $islist = 1;
    }
    elsif ($node =~ /^\s/m) {
      # verbatim
      print "<pre>$node</pre>\n";
    }
    elsif ($node =~ /^(P<[^>]*>)$/s) {
      # include another page
      my $tree = new Pod::Tree;
      $tree->load_string("P<$1>");
      $tree->walk(\&PodWiki::Runner::walker);
    }
    else {
      if ($islist) {
	# last $node was a list, but this one
	# is not, so the list is complete
	&print_list(@list);
	@list = ();
	$islist = 0;
      }
      print "<p>$node</p>\n";
    }
  }
  if (@list) {
    &print_list(@list);
  }
}

sub print_list {
  my @list = @_;
  print "<ul>\n";
  foreach my $item (@list) {
    print "<li><p>$item</p></li>\n";
  }
  print "</ul>\n";
}


1;
