 #
 #   $Id: Runner.pm,v 1.49 2005/03/24 23:02:46 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #
 #   Additional Copyrights:
 #
 #

package PodWiki::Runner;

$PodWiki::Runner::VERSION = "2.11";

use lib qw(. lib);

use strict;
no strict "refs";

use Exporter;

use Data::Dumper;
use Pod::Tree;
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use Text::Wrap qw(wrap fill);

use DB_File;

use Fcntl qw(LOCK_EX LOCK_UN);

use Rcs;
use IO::Socket;

use PodWiki::Config;
use PodWiki::Vars;
use PodWiki::Formatter;
use PodWiki::Revision;
use PodWiki::Struct;
use PodWiki::UI;
use PodWiki::Auth;

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS @docpages %head @rcs);

my (%seen, %next, %variables, %index, %comment);

use vars qw($vars $sub2go $query $me $current_user $page $input $users $groups
            $output $break_mode $config $saving $entry $width $height $attributes
	    $internal_link_marker $external_link_marker $fmttr $page_conf $page_conf_done);

@ISA=qw(Exporter PodWiki::UI PodWiki::Auth);
@EXPORT=qw();
@EXPORT_OK=qw();
@EXPORT_TAGS=();


# the syntax map pod => html
my %syntax = (
	      head1       => "h1",
	      head2       => "h2",
	      head3       => "h3",
	      head4       => "h4",
	      text        => "p",
	      verbatim    => "pre",
	      b           => "b",
	      i           => "i",
	      c           => "code",
	      f           => "i",
	      list_number => "ol",
	      list_bullet => "ul",
	      list_text   => "dl",
	      item_number => "li",
	      item_bullet => "li",
	      item_text   => "dd",
	      table       => "table",
	      row         => "tr",
	      headcell    => "th",
	      cell        => "td",
	      );



# valid POD commands
my $POD = "head1|head2|head3|head4|over|back|pod|for|begin|cut|end|item|table|tableback|row|rowback|cell|form|field";

sub new {
  my($this, %param) = @_;
  my $class = ref($this) || $this;
  my $self = \%param;
  bless($self, $class);
  return $self;
}

sub clean {
  # reset "global" VARS
  %head = %seen = %next = %variables = %index = %comment = @docpages = ();

  $config = $sub2go = $query = $me = $current_user = $width
  = $height = $page = $output = $break_mode = $saving = $entry = "";

  $page_conf = $input = $users = $groups = $attributes = $config = {};

  $page_conf_done = 0;
}

sub run {
  my($this, %param) = @_;
  $this->clean();

  $me = $this->{me};

  $config = $this->{config};
  $config->fetch();

  if ($config->{ApacheTwoHangWorkaround}) {
      STDERR->blocking(0);
  }

  if (! exists $config->{default_nonpod_formatter}) {
    $config->{default_nonpod_formatter} = 'wikishorthand';
  }

  if (! exists $config->{separator_character}) {
    $config->{separator_character} = '&sdot;';
  }

  if (! exists $config->{mark_internal_links}) {
    $config->{mark_internal_links} = 1;
  }

  if (! exists $config->{mark_external_links}) {
    $config->{mark_external_links} = 1;
  }

  if (! exists $config->{editlock_timeout}) {
    $config->{editlock_timeout} = 1800; # 30min
  }

  if (! exists $config->{rcsdebug}) {
    $config->{rcsdebug} = 0; # default: off
  }

  # prepare RCS
  @rcs = ("bindir",  $config->{rcsbindir},
	  "workdir", $config->{data},
	  "debug",   $config->{rcsdebug},
	  "file"
	 );

  if (%param) {
    # offline output
    $input = \%param;
    $config->{direct} = 1;
    $config->{mark_internal_links} = 0;
    $config->{mark_external_links} = 0;
  }
  else {
    # parse cgi input
    $query = new CGI;
    my %in = map { $_ => $query->param($_) } $query->param;
    $input = \%in;
  }

  # determine where to go
  $sub2go = $input->{state} || "";

  # POST processing of "state"
  # for binary (images) output
  if ($sub2go eq "imagethumbnail") {
    # binary output
    &imagethumbnail;
    exit;
  }

  # POST processing of "state"
  # for user preferences Cookies
  if ($sub2go eq "saveprefs") {
    my ($user, undef) = &fetch_session;
    if ($user eq $input->{prefuser}) {
      delete $input->{submit};
      delete $input->{delete};
      delete $input->{state};
      delete $input->{prefuser};
      $sub2go = "";
      &set_pref_cookie(%{$input});
      $input->{page} = "UserHome_" . $user;
    }
    # else: FIXME: print error message? if yes: HOW?
    $sub2go = "";
  }

  # fetch user prefs, if any
  &fetch_user_prefs;

  # pre-set globals
  $page   = $input->{page}   || $input->{setpage} || "PodWikiIndex";
  $entry  = $input->{entry}  || "AutoLoad";
  $height = $input->{height} || 25;
  $width  = $input->{width}  || 80;
  $page   =~ s/[`|\/\\\.\'\"]//g; # remove any unallowed characters
  # The characters at the end of this line are to get the emacs syntax highlighting back in sync: `]
  if ($config->{mark_internal_links}) {
      $internal_link_marker = "<img src=\"$config->{gfx}/ilink.png\" alt=\"Internal Link\" class=\"ilink\" />\&nbsp;";
  }
  if ($config->{mark_external_links}) {
      $external_link_marker = "<img src=\"$config->{gfx}/elink.png\" alt=\"External Link\" class=\"elink\" />\&nbsp;";
  }

  # before we go ahead, we check if authentication
  # is required
  &check_auth;

  # check if readonly is enabled and the current user (if any)
  # is allowed to override read-only state

  if ($config->{read_only} && $current_user) {
    if (grep { $_ eq $current_user } split /\s*,\s*/, $config->{read_only_override_users}) {
      $config->{read_only} = 0;
    }
  }

  # load vars
  $vars = new PodWiki::Vars;

  # Get the formatter object
  $fmttr = PodWiki::Formatter->new("config" => $config);

  if (! %param) {
    # output is html
    print "Content-Type: text/html\n\n\n";
  }
  else {
    # offline output
    $config->{direct} = 1;
    $config->{mark_internal_links} = 0;
    $config->{mark_external_links} = 0;
  }

  eval {
    if (&data_empty) {
      # doing the initial setup
      $sub2go = "change_root_passwd";
      &setup;
    }
    &display;
    &detach;
  };
  if ($@) {
    die $@;
  }
}

sub pod_link {
    my ($this, $linktext) = @_;
    my $text = "";

    my($link, $name, $target, $uri);
    $link = $name = $target = $uri = "";

    ($name, $target) = split /\|/, $linktext;
    if ($target) {
	$uri = $target;
    } else {
	$name = $linktext;
	$uri = $linktext;
    }

    # FIXME: This code essentially duplicates that in the walk_sequence
    #code for a link, except that it returns the formatted link as a
    #string, rather than printing it to STDOUT. At some point, it would
    #be nice to make these the same code.  

    my $section      = "";

    if ($uri =~ /^(http|https|ftp):\/\/(.*)/) {
	# external uri
	my $ziel = $2;
	my $marker = $external_link_marker;
	if ($ziel =~ /^[\/\.]/) {
	    # local path, remove http:// part
	    $uri = $ziel;
	    $marker = $internal_link_marker; # Change to internal link marker
	}
	$text .= "${marker}<a href=\"$uri";
    }
    elsif ($uri =~ /^mailto:(.*)$/) {
	# mail uri
	# FIXME: fragment addresse here?!
	my $address = $1;
	$text .= "${external_link_marker}<a href=\"mailto:$address";
    }
    elsif ($uri =~ /^(state|intern):(.*)$/) {
	# internal state uri
	my ($state, $cgi_params) = split/\?/, $2;
	$text .= "${internal_link_marker}<a href=\"$me?state=" . lc($state);
	if ($cgi_params) {
	    $text .= "&$cgi_params";
	}
    }
    else {
	# internal uri (podwiki word)
	if ($uri =~ m/(.*?)\#(.*)/) {
	    $uri = $1;
	    $section = $2;
	}
	if (-e "$config->{data}/$uri") {
	    # URI exists
	    $text .= "${internal_link_marker}<a href=\"";
	    if (exists $input->{linkhtml}) {
	      $uri = "${uri}.html";
	    }
	    if ($config->{direct}) {
		$text .= "$uri";
	    }
	    else {
		$text .= "$me?page=$uri";
	    }
	}
	else {
	    # nope
	    $name = "?";
	    my $class;
	    if (-e "$config->{data}/.attic/$uri") {
		$class = qq(style="background: red;");
	    }
	    else {
		$class = qq(class="nolink");
	    }
	    $text .= "<font $class>$uri</font><a href=\"$me?page=$uri&state=edit&inherit=$page";
	}
    }
    if ($section) {
	# anchor reference
	$text .= "#" . $section;
    }
    $text .= "\">";
    if ($name) {
	$text .= $name;
    }
    $text .= "</a>";
    return $text;
}

sub display {
  # the core of the engine, fetch the page AutoLoad,
  # which itself includes all the stuff which drives
  # the piece.
  &tie_attributes($page);
  my $msg = "";

  if (exists $input->{unlock}) {
    # forced UNLOCK
    my $rcs = new PodWiki::Revision(@rcs, "$page");
    $rcs->UNLOCK();
    delete $input->{unlock};
    $msg = qq(<p class="nolink">UNLOCKED page $page</p>\n);
  }

  &render(&fetch_page($entry));
  print $output;
  print $msg;
}

sub src {
  if (-e "$config->{data}/$page") {
    my $content = &fetch_page($page);
    $content =~ s/</&lt;/g;
    $content =~ s/>/&gt;/g;
    $content =~ s/&nbsp;/&amp;nbsp;/g;
    $Text::Wrap::columns = 72;
    $content =  fill("", "", split /\n/, $content);
    print qq(<h3>Page Source for $page</h3><pre>$content</pre>);
  }
}



sub edit {
  my ($content, $revision, $edit_title, $inherit, $attribute_source);

  return if $config->{read_only};

  $inherit = $content = "";
  if ($input->{preview}) {
    $content = $input->{content};
    ##### $content =~ s/\r//g;
    print "<h3>Page Preview:</h3><div class=\"pdotted\">\n";
    &render($content);
    print "</div>";
  }
  else {
    if (-e "$config->{data}/$page") {
      $content = &fetch_page($page);
    }
  }

  if (-e "$config->{data}/$page") {
    my $rcs = new PodWiki::Revision(@rcs, "$page");

    if ($rcs->lock() && ! $input->{preview}) {
      # page is locked
      my $lockuser = $rcs->lockuser();

      # check if we can timeout the lock
      my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	  $atime,$mtime,$ctime,$blksize,$blocks)
	= stat("$config->{data}/$page");
      my $diff = time - $mtime;
      if ($diff <= $config->{editlock_timeout} ) {
	# inside 1 hour timeout
	my $pending = int(($config->{editlock_timeout} - $diff) / 60);
	print qq(<p class="nolink">Page is LOCKED and currently being edited by user <b>$lockuser</b>.
                  <br />UNLOCK Timeout will arrise in $pending minutes.
                 </p>\n);
	print qq(<p class="pdotted">You may override the lock if you know what you are doing:
                 <a href="$me?page=$page&unlock=1&state=edit">OVERRIDE LOCK</a></p>\n);
	return;
      }
      else {
	$rcs->UNLOCK();
      }
    }

    if (! $input->{preview}) {
      # well, LOCK the page now, if in preview mode, it is already locked
      my $user = $current_user || "anonymous";
      $rcs->LOCK($user);
    }

    $revision = $rcs->head;
    $edit_title = qq(Edit Page: ${internal_link_marker}<i><a href="$me?page=$page">$page</a></i>);
    $attribute_source = $page;
  }
  else {
    if (-e "$config->{data}/.attic/$page") {
      print qq(
        <div class="pdotted">
          <h4>CAUTION!</h4>
          <p>
           This page has been removed some time ago. A backup copy exists.<br />
           Do you want to restore the file?
           <a href="$me?state=do_restore_attic_admin&pages=$page">Restore \"$page\" from .attic/</a>
          </p>
        </div>
      );
    }
    $revision = "NEW";
    $edit_title = qq(Create Page: <input type="text" name="setpage" value="$page">);
    if (exists $input->{inherit}) {
      # add note for editor
      $inherit = qq( properties inherited by
		     ${internal_link_marker}<a href="$me?page=$input->{inherit}">$input->{inherit}</a>);

      # where to fetch attributes from
      $attribute_source = $input->{inherit};

      # use parent options, if any
      my $page_conf = &parse_options($attribute_source);
      if (scalar keys %{$page_conf}) {
	# add this to the content
	$content  = "=begin options\n\n";
	$content .= join "\n", map { "$_ = $page_conf->{$_}" } sort keys %{$page_conf};
	$content .= "\n\n=end options\n";
      }
    }
  }

  $content =~ s/&([a-zA-Z]+);/&amp;$1;/g;
  $content =~ s/</&lt;/g;
  $content =~ s/>/&gt;/g;

  print qq(
<form title="PodWiki" action="$me" method="post">
<h3>$edit_title (Revision: $revision)</h3>
$inherit
<p>Get some help about );
  print PodWiki::Runner->pod_link("PodWikiMarkup") . " or " . PodWiki::Runner->pod_link("WikiShorthand") . "</p>\n";
  print qq (
 <input type="hidden" name="page" value="$page">
 <input type="hidden" name="state" value="save">
 <textarea id="content" name="content" rows="$height" cols="$width">$content</textarea>
 <a href="javascript:ewiki_enlarge()" style="text-decoration:none">+</a>
 <script type="text/javascript">
   <!--
     function ewiki_enlarge() {
       var ta=document.getElementById("content");
       ta.style.width=((ta.cols*=1.0)*10).toString()+"px";
       ta.style.height=((ta.rows*=1.1)*20).toString()+"px";
     }
   //-->
 </script>
 <br />
Comment:<br />
<input type="text" name="comment" class="dotted" size="70"><br /><br />
State: <select name="rcs_state">
<option selected>Experimental</option>
<option>Testing</option>
<option>Bugfix</option>
<option>Final</option>
</select>
);


  my $a_size = scalar keys %{$attributes};
  if ($a_size) {
    # attributes exists, so provide them
    print "<h4>Classify page:</h4>";
    print qq(<table border="0"><tr valign="top">);
    my $sel;
    foreach my $attribute (sort keys %{$attributes}) {
      next if $attribute =~ /^\./;
      print qq(<td><b>$attribute</b><br />);
      print qq(<select name="podattributes-${attribute}" size="10"><option></option>);
      foreach my $attr_value (sort split /\s\s*/, $attributes->{$attribute}) {
	if (exists $attributes->{".classified"}->{$attribute}->{$attr_value}->{$attribute_source}) {
	  $sel = " selected";
	}
	else {
	  $sel = "";
	}
	print qq(<option${sel}>$attr_value</option>);
      }
      print qq(</select></td>);
    }
    print qq(</tr></table>);
  }

  print qq(
 <input type="submit" name="save" value=" &nbsp; Save &nbsp; ">
 <input type="submit" name="preview" value=" &nbsp; Preview &nbsp; ">
 <a href="$me?page=$page&unlock=1">Cancel</a>
 </form>
 <hr noshade>
 <h3>Upload an image or another file:</h3>
 <form title="upload" action="$me" method="post" enctype="multipart/form-data" >
   <input type=hidden name=state  value="upload">
   <input type=hidden name=page value="$page">
   <input type=file name=file size=40>
   <input type=submit name=submit value="upload this file">
 </form>

 <p><a href="$me?state=fileindex">Manage uploaded files (fileindex)</a></p>
  );
}


sub edit_bar {
  my($sep) = $config->{separator_character} || "";

  print qq(<div class="editbar">);

  if ($current_user) {
    print qq(<font size="-2">logged in as
       <b><a href="$me?state=change_passwd">$current_user</a></b>
       </font>
       $sep <a class="dotted" href="$me?state=logout">logout</a>);
  }
  else {
    print qq(<a class="dotted" href="$me?state=login">login</a>);
  }

  if (! $config->{read_only}) {
    if (! $sub2go || ($sub2go eq "save" && -e "$config->{data}/$page")) {
      my $rcs = new PodWiki::Revision(@rcs, "$page");
      my $revision = $rcs->head;
      my $modified = scalar localtime($rcs->revdate);
      print qq(
      $sep <font size="-2"><i>Last modified: $modified. Revision: $revision</i></font>
      $sep <a class="dotted" href="$me?page=$page&state=edit">EditThisPage</a>
      $sep <a class="dotted" href="$me?page=$page&state=src">ShowSource</a>
      $sep <a class="dotted" href="$me?page=$page&state=info">InfoAboutPage</a>
    );
    }

    if (&is_member($current_user, "root")) {
      print qq( $sep <a class="dotted" href="$me?state=site_admin">SiteAdmin</a>);
    }
  }

  print qq(</div>);
}


sub site_admin {
  return if $config->{read_only};

  print qq(
<h3>Welcome to the <I>PodWiki</I> Site-Admin Page</h3>

<table width="100%" border="0" cellspacing="10"><tr valign="top"><td >

<h4>Users</h4>
<form class="pdotted" title="newuser" action="$me" method="post">
<input type="hidden" name="state" value="new_user_admin">
New user: <input type="text" size="20" name="user">
<input type="submit" name="submit" value="create">
</form>
  );

  &tie_users;
  print qq(

 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>User</th><th>Action</th>
   </tr>
  );

  foreach my $user (sort keys %{$users}) {
    my $home = "UserHome_${user}";
    my $link;
    if (-e "$config->{data}/$home") {
#      $link = qq(${internal_link_marker}<a href="$me?page=$home">$user</a>);
      $link = PodWiki::Runner->pod_link("${user}|${home}");
    }
    else {
      $link = qq(<font class="nolink">$user</font><a href="$me?page=$home&state=edit">?</a>);
    }
    print qq(
    <tr bgcolor="#ffffff">
      <td>$link</td>
      <td>
        <a class="dotted" href="$me?page=$page&state=edit_user_admin&user=$user">edit</a>
        <a class="dotted" href="$me?page=$page&state=remove_user_admin&user=$user">remove</a>
      </td>
    </tr>
    );
  }

  print qq(</table></td></tr></table><p>&nbsp;</p>);



  print qq(
</td><td bgcolor="#ccccff"></td><td>

<h4>Groups</h4>
<form class="pdotted" title="newgroup" action="$me" method="post">
<input type="hidden" name="state" value="new_group_admin">
New group: <input type="text" size="20" name="group">
<input type="submit" name="submit" value="create">
</form>
  );

  &tie_groups;
  print qq(
 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>Group</th><th>Action</th>
   </tr>
  );

  foreach my $group (sort keys %{$groups}) {
    print qq(
    <tr bgcolor="#ffffff">
      <td>$group</td>
      <td>
         <a class="dotted" href="$me?page=$page&state=edit_group_admin&group=$group">edit</a>
         <a class="dotted" href="$me?page=$page&state=remove_group_admin&group=$group">remove</a>
      </td>
    </tr>
    );
  }

  print qq(</table></td></tr></table></td></tr></table>);

  print qq(<h4>Attributes</h4>
<form class="pdotted" title="newattribute" action="$me" method="post">
<input type="hidden" name="state" value="new_attribute_admin">
New attribute: <input type="text" size="20" name="attribute">
<input type="submit" name="submit" value="create">
</form>
  );

  print qq(
 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>Attribute</th><th>Values (separated by whitespace)</th><th>Action</th>
   </tr>
  );

  foreach my $attribute ( sort keys %{$attributes}) {
    next if $attribute =~ /^\./;
    print qq(
      <form title="change_${attribute}" action="$me" method="post">
      <input type="hidden" name="state" value="edit_attribute_admin">
      <input type="hidden" name="attribute" value="$attribute">
      <tr>
        <td width="10%" bgcolor="#ffffff"><b>$attribute</b></td>
        <td width="70%" bgcolor="#ffffff">
         <input type="text" name="attr_values" size="60" value="$attributes->{$attribute}">
        </td>
        <td bgcolor="#ffffff">
          <input type="submit" name="submit" value="change">
          <a class="dotted" href="$me?page=$page&state=remove_attribute_admin&attribute=$attribute">remove</a>
        </td>
      </tr>
      </form>
      );
    }

  print qq(</table></td></tr></table>);

  &tie_variables;
    print qq(<h4>Variables</h4>
<form class="pdotted" title="newvariable" action="$me" method="post">
<input type="hidden" name="state" value="new_variable_admin">
New variable: <input type="text" size="20" name="variable">
<input type="submit" name="submit" value="create">
</form>
  );

    print qq(
 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>Variable</th><th>Value</th><th>Action</th>
   </tr>
  );

  foreach my $var (sort keys %variables) {
        print qq(
      <form title="change_${var}" action="$me" method="post">
      <input type="hidden" name="state" value="edit_variable_admin">
      <input type="hidden" name="variable" value="$var">
      <tr>
        <td width="10%" bgcolor="#ffffff"><b>\$$var</b></td>
        <td width="70%" bgcolor="#ffffff">
         <input type="text" name="var_value" size="60" value="$variables{$var}">
        </td>
        <td bgcolor="#ffffff">
          <input type="submit" name="submit" value="change">
          <a class="dotted" href="$me?page=$page&state=remove_variable_admin&variable=$var">remove</a>
        </td>
      </tr>
      </form>
      );
  }

  print qq(</table></td></tr></table>);

  print qq(
    <p>&nbsp;</p>
    <h4>File Maintenance</h4>
    <ul>
      <li><p><a href="$me?state=docupdate_admin">Update PodWiki Pages and Documentation</a></p></li>
      <li><p><a href="$me?state=restore_attic_admin">Restore removed pages and files</a></p></li>
    </ul>
  );
}

sub docupdate_admin {
  return if $config->{read_only};

  print qq(
     <h3>Update PodWiki Pages and Documentation</h3>
     <form title="docupdate" method="post" action="$me">
        <input type="hidden" name="state" value="do_docupdate_admin">
	<select name="pages" size="10" multiple>
  );

  opendir D, "$config->{templates}" or die "Could not read DATA directory $config->{data}: $!\n";
  foreach my $docpage (sort grep { $_ !~ /^\./ } readdir(D)) {
    print qq(<option selected>$docpage</option>);
  }
  closedir D;

  print qq(
	</select>
      <br /><input type="submit" name="submit" value="Update Pages from templates/ directory">
     </form>
  );
}

sub do_docupdate_admin {
  return if $config->{read_only};

  my @docpages = $query->param("pages");
  foreach my $file (@docpages) {
    my $rcs = new PodWiki::Revision(@rcs, "$file");
    $rcs->LOCK();
    system("cp '$config->{templates}/$file' '$config->{data}/$file'") and
      die "Could not install template file $file: $!\n";
    $rcs->commit("UpdatedByPodWikiSetup", "PodWiki(root)");
  }
  print qq(<p class="nolink">PodWiki documentation has been updated.</p>);
}

sub new_attribute_admin {
  return if $config->{read_only};
  if (! exists $attributes->{$input->{attribute}}) {
    # create it
    $attributes->{$input->{attribute}} = "";
    PodWiki::Struct::store($attributes, "$config->{data}/.podattributes");
  }
  else {
    print qq(<p class="nolink">Attribute <b>$input->{attribute}</b> already exists!</p>);
  }
  &site_admin;
}

sub remove_attribute_admin {
  return if $config->{read_only};
  if (! exists $attributes->{$input->{attribute}}) {
    print qq(<p class="nolink">Attribute <b>$input->{attribute}</b> does not exist!</p>);
  }
  else {
    delete $attributes->{$input->{attribute}};
    PodWiki::Struct::store($attributes, "$config->{data}/.podattributes");
    print qq(<p class="nolink">Attribute <b>$input->{attribute}</b> has been removed.</p>);
  }
  &site_admin;
}

sub edit_attribute_admin {
  return if $config->{read_only};
  if (! exists $attributes->{$input->{attribute}}) {
    print qq(<p class="nolink">Attribute <b>$input->{attribute}</b> does not exist!</p>);
  }
  else {
    my $values = join " ", sort split /\s\s*/, $input->{attr_values};
    $attributes->{$input->{attribute}} = $values;
    PodWiki::Struct::store($attributes, "$config->{data}/.podattributes");
    print qq(<p class="nolink">Attribute <b>$input->{attribute}</b> has been changed.</p>);
  }
  &site_admin;
}

sub new_variable_admin {
  return if $config->{read_only};
  if (! exists $variables{$input->{variable}}) {
    # create it
    $variables{$input->{variable}} = "";
    PodWiki::Struct::store(\%variables, "$config->{data}/.podvariables");
  }
  else {
    print qq(<p class="nolink">Variable <b>$input->{variable}</b> already exists!</p>);
  }
  &site_admin;
}

sub remove_variable_admin {
  return if $config->{read_only};
  if (! exists $variables{$input->{variable}}) {
    print qq(<p class="nolink">Variable <b>$input->{variable}</b> does not exist!</p>);
  }
  else {
    delete $variables{$input->{variable}};
    PodWiki::Struct::store(\%variables, "$config->{data}/.podvariables");
    print qq(<p class="nolink">Variable <b>$input->{variable}</b> has been removed.</p>);
  }
  &site_admin;
}

sub edit_variable_admin {
  return if $config->{read_only};
  if (! exists $variables{$input->{variable}}) {
    print qq(<p class="nolink">Variable <b>$input->{variable}</b> does not exist!</p>);
  }
  else {
    $variables{$input->{variable}} = $input->{var_value};
    PodWiki::Struct::store(\%variables, "$config->{data}/.podvariables");
    print qq(<p class="nolink">Variable <b>$input->{variable}</b> has been changed.</p>);
  }
  &site_admin;
}

sub edit_user_admin {
  return if $config->{read_only};
  my($new) = @_;
  my $ro = "";
  if (! $new) {
    $ro = "readonly";
  }
  else {
    # new user
    &tie_users;
    if (exists $users->{$input->{user}}) {
      print qq(<p class="nolink">User <b>$input->{user} already exists. Please try another username.</p>);
      $input->{user} = "";
    }
  }
  my($user, $pw, $mail) = &fetch_user($input->{user});
  print qq(
   <h3>Edit user $input->{user}</h3>
   <form title="edituser" action="$me" method="post">
   <input type="hidden" name="state" value="do_save_user_admin">
    <table border="0">
      <tr>
        <th align="right">User</th>
	<td><input type="text" size="20" name="edituser" value="$input->{user}" $ro></td>
      </tr>
      <tr>
        <th align="right">Email</th>
	<td><input type="text" size="20" name="mail" value="$mail"></td>
      </tr>
      <tr>
        <th align="right">New password</th>
	<td><input type="password" size="20" name="editpasswd"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save"></td>
      </tr>
    </table>
   </form>
  );
}

sub new_user_admin {
  return if $config->{read_only};
  &edit_user_admin("new");
}

sub do_save_user_admin {
  return if $config->{read_only};
  if ($input->{edituser} =~ /[`|\/\\\.\'\"]/) {
    # The characters at the end of this line are to get the emacs syntax highlighting back in sync: `]
    &edit_user_admin("new");
    print qq(<p class="nolink">Username contained forbidden characters. Please try another username.</p>);
    return;
  }

  if (($input->{editpasswd} || $input->{mail}) && $input->{edituser}) {
    &save_user($input->{edituser}, $input->{editpasswd}, $input->{mail});
    &site_admin;
    print qq(<p class="nolink">User <b>$input->{user}</b> saved</p>);
    return;
  }
  else {
    &site_admin;
    print qq(<p class="nolink">Incomplete input</p>);
    return;
  }
}

sub remove_user_admin {
  return if $config->{read_only};
  if ($input->{user}) {
    &del_user($input->{user});
    &del_member($input->{user}, "*");
    &site_admin;
    print qq(<p class="nolink">User <b>$input->{user}</b> removed</p>);
  }
}


sub edit_group_admin {
  return if $config->{read_only};
  my($new) = @_;
  my $ro;
  if (! $new) {
    $ro = "readonly";
  }

  print qq(
   <h3>Edit group $input->{group}</h3>
   <form title="editgroup" action="$me" method="post">
   <input type="hidden" name="state" value="do_save_group_admin">
   <input type="hidden" name="group" value="$input->{group}" $ro>
    <table border="0">
      <tr>
        <th align="right">Group</th>
	<td><input type="text" name="group" value="$input->{group}" $ro></td>
      </tr>
      <tr valign="top">
        <th align="right">Members</th>
	<td>
	  <select name="members" size="10" multiple>
    );

    &tie_users;
    foreach my $user (sort keys %{$users}) {
      next if(! $user);
      if (! $new) {
	if (&is_member($user, $input->{group})) {
	  print qq(      <option selected>$user</option>);
	}
	else {
	  print qq(      <option>$user</option>);
	}
      }
      else {
	print qq(      <option>$user</option>);
      }
    }

    print qq(	  </select>
      <br /><font size="-1">Select multiple users by pressing CTRL while clicking on user names</font>
        </td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save"></td>
      </tr>
    </table>
   </form>
  );
}

sub remove_group_admin {
  return if $config->{read_only};
  if ($input->{group}) {
    &del_group($input->{group});
    &site_admin;
    print qq(<p class="nolink">Group <b>$input->{group}</b> removed</p>);
  }
}

sub new_group_admin {
  return if $config->{read_only};
  &edit_group_admin("new");
}

sub do_save_group_admin {
  return if $config->{read_only};
  my @members = $query->param("members");  # shall be an array
  if ($input->{group}) {
    &set_members($input->{group}, @members); # creates the group
    &site_admin;
    print qq(<p class="nolink">Group <b>$input->{group}</b> saved</p>);
  }
  else {
    &site_admin;
    print qq(<p class="nolink">Incomplete input</p>);
    return;
  }
}


sub clean_head {
  my $head = shift;
  $head =~ s/[><\.\_\-\(\)\/\\\#\:\;\"\'\s]//g;
  return $head;
}

sub fetch_toc {
  my ($content, $page) = @_;
  local $_;
  @{$head{$page}} = ();
  foreach (split /\n/, $content) {
    chomp;
    if (/^=head(\d) (.*$)$/) {
      my $level  = $1;
      my $target = $2;
      push @{$head{$page}}, {
		   level  => $level,
		   target => &clean_head($target),
		   title  => $target
		   };
    }
  }
}

sub fetch_page {
  my $page = shift;
  if (exists $seen{$page} && ($sub2go ne "edit" && $sub2go ne "src")) {
    # this content has already been displayed
    return "";
  }
  else {
    $seen{$page} = 1;
    open T, "<$config->{data}/$page" or die "Could not read page $page: $!\n";
    my $content = join "", <T>;
    close T;
    &fetch_toc($content, $page);
    return $content;
  }
}

sub process_template {
  #
  # replace the D<> and M<> template data fields
  # with the given values in %input
  my $tpl = &fetch_page($input->{template});
  my $content = "";
  if ($tpl =~ /(=form.*=formback)/s) {
    my $form = "$1";

    $form =~ s/=formback//g;
    $form =~ s/=form//g;

    $form =~ s/(D|M)<(.+)>/$input->{$2}/g;
    $content .= $form;
  }
  $tpl =~ s/=form.+=formback/$content/s;

  $tpl = qq(<!-- created from template '$input->{template}' -->\n\n) . $tpl;
  return $tpl;
}

sub save {
  return if $config->{read_only};
  if ($input->{preview}) {
    &edit;
    return;
  }
  my $content = "";

  if (exists $input->{template}) {
    $content = &process_template();
  }
  else {
    $content = $input->{content};
  }

  if (! $input->{page} && ! $content) {
    # nothing to be done
    print qq(<p class="nolink">No content, nothing saved.</p>);
    return;
  }

  if (exists $input->{setpage}) {
    # newly created page
    if (exists $input->{template}) {
      if(! $input->{setpage}) {
	$input->{setpage} = &getrandom(16);
      }
    }
    if (-e "$config->{data}/$input->{setpage}") {
      print qq(<p class="nolink">Page <b>$input->{setpage}</b> already exists!</p>);
      return;
    }
    else {
      $page = $input->{setpage};
    }
  }

  if (-e "$config->{data}/$page" && ! -w "$config->{data}/$page") {
    print qq(<p class="nolink">Could not save page <b>$page</b>. LOCK has been overridden by another user</p>\n);
    return;
  }

  my $lock = &lock_podwiki();

  my $rcs = new PodWiki::Revision(@rcs, "$page");

  my $file = &untaint("$config->{data}/$page");
  open T, ">$file" or die "Could not write page $page: $!\n";
  print T $content;
  if ($content !~ /\n$/s) {
    print T "\n";
  }
  close T;

  if (! -e "$config->{data}/${page},v") {
    # initial checkin
    $rcs->add("InitialCheckinByPodWiki");
    $rcs->checkout();
  }
  else {
    my $user;
    $input->{comment} ||= "*** empty log message ***";
    $input->{login} = $current_user || "anonymous";
    $input->{login} .= "-" . $ENV{REMOTE_ADDR};

    $rcs->commit("$input->{comment}", "$input->{login}", "$input->{rcs_state}");
  }

  # re-build the page index based on X<> indices
  $saving = 1;
  &tie_index;
  &render(&fetch_page($page));

  PodWiki::Struct::store(\%index, "$config->{data}/.podindex");

  foreach my $attribute (keys %{$attributes}) {
    # look for classification
    if (exists $input->{"podattributes-${attribute}"}) {
      # ok, defined, classify the page
      &store_attribute($attribute, $input->{"podattributes-${attribute}"});
      &tie_attributes;# also retrieve after saving to avoid dead-locks
    }
  }

  print qq(<p class="nolink">Saved successfully page <a href="$me?page=$page"><b>$page</b></a>.</p>);

  &unlock_podwiki($lock);
}

sub interpolate {
  # interpolate embedded variables, if any
  my $content = shift;

  # interpolate internal variables
  $content =~ s/(?<!\\)\$page/$page/g;
  $content =~ s/(?<!\\)\$me/$me/g;
  $content =~ s/(?<!\\)\$version/$PodWiki::VERSION/g;
  $content =~ s/(?<!\\)\$entry/$entry/g;
  $content =~ s/(?<!\\)\$height/$height/g;
  $content =~ s/(?<!\\)\$width/$width/g;
  $content =~ s/(?<!\\)\$sep/$config->{separator_character}/g;

  # interpolate global variables
  &tie_variables;
  if (scalar keys %variables > 0) {
    my $pref = join "|", keys %variables;
    $content =~ s/(?<!\\)(\$)($pref)/$variables{${2}}/g;
  }

  # interpolate user defined variables in =begin options block
  if (scalar keys %{$page_conf} > 0) {
    my $pref = join "|", keys %{$page_conf};
    $content =~ s/(?<!\\)(\$)($pref)/$page_conf->{${2}}/g;
  }

  # interpolate PodWiki::Vars variables
  $content = $vars->interpolate($content);

  # interpolate masked \$ characters to $
  $content =~ s/\\\$/\$/g;

  # remove \r
  $content =~ s/\r//g;

  return $content;
}

sub listvars {
    # Internal state: List defined variables
    my $content = "";

    # start table
    $content .= "\n\n<table border=\"1\">\n";
    $content .= "<thead>\n";
    $content .= "<tr><th>Variable</th><th>Description</th><th>Type</th><th>Value</th></tr>\n";
    $content .= "</thead><tbody>\n";

    # list internal variables
    my %intvars = ( page => [$page, "Name of the current page"],
		    me => [$me, "Name of the PodWiki script"],
		    version => [$PodWiki::VERSION, "PodWiki version"],
		    entry => [$entry, "Name of first page loaded"],
		    height => [$height, "Height of textarea used for editing pages"],
		    width => [$width, "Width of textarea used for editing pages"],
		    sep => [$config->{separator_character}, "Separator character (used in, for example, horizontal menus)"],
		    );

    foreach my $key (sort keys %intvars) {
	$content .= "<tr><td> " . $key . " </td><td>" . ${$intvars{$key}}[1] . "</td><td>Internal</td><td>" . ${$intvars{$key}}[0] . "</td></tr>\n";
    }

    # list global variables
    &tie_variables;
    foreach my $key (sort keys %variables) {
	$content .= "<tr><td> " . $key . " </td><td>" . "</td><td>Global</td><td>" . $variables{$key} . "</td></tr>\n";
    }

    # list user defined variables in =begin options block
    foreach my $key (sort keys %{$page_conf}) {
	$content .= "<tr><td> " . $key . " </td><td>" . "</td><td>Page</td><td>" . $page_conf->{$key} . "</td></tr>\n";
    }

    # list PodWiki::Vars variables
    $content .= $vars->listvars();


    # close table
    $content .= "</tbody></table>\n\n";

    # render the result
    &render($content);
    return $content;
}


sub comment {
  &tie_comment;
  my $mode = $input->{mode} || "";

  if ($mode eq "add") {
    &add_comment();
  }
  elsif ($mode eq "save") {
    # save a posted comment
    if (exists $input->{preview}) {
      my $reply_id = $input->{reply_id} || "";
      &add_comment($reply_id);
      return;
    }
    my $max = (sort {$a <=> $b} keys %{$comment{$page}})[(scalar keys %{$comment{$page}}) -1];
    my $id  = $max + 1;
    my %h = (
	     time    => time,
	     user    => $input->{user},
	     subject => $input->{subject},
	     comment => $input->{comment},
	     id      => $id
	     );
    $comment{$page}->{$id} = \%h;

    if (exists $input->{reply_id}) {
      # register this comment as reply to comment $reply_id
      $comment{$page}->{$input->{reply_id}}->{replies}->{$id} = 1;
    }

    PodWiki::Struct::store(\%comment, "$config->{data}/.podcomments");

    delete $input->{mode}; # avoid re-posting of comment during P<state:comment>
    &render(&fetch_page($page));
    print qq(<p class="nolink">Comment $id successfully saved.</p>);
  }
  elsif ($mode eq "reply") {
    # present "add comment" form along with reference to an existing comment
    &add_comment($input->{id});
  }
  elsif ($mode eq "remove") {
    delete $comment{$page}->{$input->{id}};

    PodWiki::Struct::store(\%comment, "$config->{data}/.podcomments");

    delete $input->{mode};
    &render(&fetch_page($page));
    print qq(<p class="nolink">Comment $input->{id} successfully removed.</p>);
  }
  else {
    # display current comments along with "add comment" link
    print qq(<hr noshade size="1">
             <h3>Comments (${internal_link_marker}<a href="$me?state=comment&mode=add&page=$page">Add
                 Comment to this Page</a>)</h3>);

    print qq(<ul style="list-style-type: none">\n);
    my %got;
    foreach my $id (sort keys %{$comment{$page}}) {
      next if exists $got{$id};
      &print_comment($comment{$page}->{$id}, $id);
      $got{$id} = 1;

      if (scalar keys %{$comment{$page}->{$id}->{replies}} >= 1) {
        print qq(<ul style="list-style-type: none">\n);
        foreach my $reply_id (sort keys %{$comment{$page}->{$id}->{replies}}) {
          if (exists $comment{$page}->{$reply_id}) {
	    next if exists $got{$reply_id};
            &print_comment($comment{$page}->{$reply_id}, $reply_id);
            $got{$reply_id} = 1;

	    if (scalar keys %{$comment{$page}->{$reply_id}->{replies}} >= 1) {
	      print qq(<ul style="list-style-type: none">\n);
	      foreach my $re_reply_id (sort keys %{$comment{$page}->{$reply_id}->{replies}}) {
		if (exists $comment{$page}->{$re_reply_id}) {
		  next if exists $got{$re_reply_id};
		  &print_comment($comment{$page}->{$re_reply_id});
		  $got{$re_reply_id} = 1;
		}
		else {
		  delete $comment{$page}->{$re_reply_id};
		}
	      }
	      print qq(</ul>\n);
	    }

          }
          else {
            delete $comment{$page}->{$reply_id};
          }
        }
        print qq(</ul>\n);
      }

    }
    print qq(</ul>\n);

    PodWiki::Struct::store(\%comment, "$config->{data}/.podcomments");
  }
}




sub add_comment {
  my $reply_id = shift;
  my $comment = "";
  my $subject = "";

  if ($input->{preview}) {
    print qq(<h3>Comment Preview:</h3><div class="pdotted">);
    &render($input->{comment});
    print qq(</div>);
    $subject = $input->{subject};
    $comment = $input->{comment};
  }

  # present "add comment" form
  print qq(
    <h3>Add comment to page );
  print PodWiki::Runner->pod_link("<i>$page</i>|$page");
  print qq(</h3>
    <form title="comment" action="$me" method="post">
    <input type="hidden" name="page" value="$page">
    <input type="hidden" name="mode" value="save">
    <input type="hidden" name="state" value="comment">
  );

  if ($reply_id) {
    print qq(<input type="hidden" name="reply_id" value="$reply_id">);
    $subject = $subject || "RE: $comment{$page}->{$reply_id}->{subject}";
  }

  my $user = $current_user || "anonymous";
  print qq(
      <table border="0">
        <tr>
	  <th align="right">User</th>
	  <td align="left"><input type="text" name="user" value="$user"</td>
	</tr>
	<tr>
	  <th align="right">Subject</th>
	  <td align="left"><input type="text" name="subject" size="60" value="$subject"</td>
	</tr>
	<tr>
	  <th align="right" valign="top">Comment</th>
	  <td align="left">
           <p>Get some help about );
  print PodWiki::Runner->pod_link("PodWikiMarkup") . " or " . PodWiki::Runner->pod_link("WikiShorthand") . "</p>\n";
  print qq(
	   <textarea name="comment" rows="10" cols="60">$comment</textarea>
	  </td>
	</tr>
	<tr>
	  <td rowspan="2">
	    <input type="submit" name="submit" value="&nbsp; Save &nbsp;">
            <input type="submit" name="preview" value="&nbsp; Preview &nbsp;">
	  </td>
	</tr>
      </table>
    </form>
    );
}

sub print_comment {
  my ($hash, $id) = @_;
  my %C = %{$hash};
  my $date = scalar localtime($C{time});
  my ($home, $reply, $remove);

  if ($id) {
    # provide reply link
    $reply = qq(<a href="$me?state=comment&mode=reply&id=$id&page=$page">reply</a>);
  }
  else {
    $reply = "";
  }

  if (-e "$config->{data}/UserHome_$C{user}") {
    $home = qq(<a href="$me?page=UserHome_$C{user}">$C{user}</a>);
  }
  else {
    $home = "<i>$C{user}</i>";
  }

  if ($current_user eq "root") {
    $remove = qq(<a href="$me?state=comment&mode=remove&id=$C{id}&page=$page">(remove comment)</a>);
  }
  else {
    $remove = "";
  }

  print qq(<li><p class="comment"><b>$C{subject}</b> by $home ($date) $reply $remove</p>\n);
  &render($C{comment});
  print qq(\n</li><p>);
}

sub list_userhomepages {
  &tie_users;
  print "<ul>\n";
  foreach my $user (sort keys %{$users}) {
    my $home = "UserHome_${user}";
    my $link;
    if (-e "$config->{data}/$home") {
      $link = qq(${internal_link_marker}<a href="$me?page=$home">$user</a>);
    }
    else {
      $link = qq(<font class="nolink">$user</font><a href="$me?page=$home&state=edit">?</a>);
    }
    print "<li>$link</li>\n";
  }
  print "</ul>\n";
}

sub remove {
  return if $config->{read_only};
  print qq(<h3>Remove</h3>);
  if ($input->{file}) {
    # image
    print qq(<p>Are you sure to remove the file <b>$input->{file}</b>?</p>);
  }
  else {
    print qq(<p>Are you sure to remove the page <b>$page?</b></p>);
  }
  print qq(
   <form title="remove" action="$me" method="post">
    <input type="hidden" name="state" value="doremove">
    <input type="hidden" name="file" value="$input->{file}">
    <input type="hidden" name="page" value="$page">
    <input type="submit" name="submit" value="YES">
    <input type="submit" name="submit" value="NO">
   </form>
  );
}

sub doremove {
  return if $config->{read_only};
  if ($input->{submit} eq "NO") {
    # cancel
    &pageindex;
  }
  else {
    if ($input->{file}) {
      # image
      &attic($config->{gfx}, $input->{file});
      unlink &untaint("$config->{gfx}/$input->{file}") or die "Could not remove file $input->{file}: $!\n";
      &fileindex;
      print qq(<p class="nolink">Removed successfully file <b>$input->{file}</b>.</p>);
    }
    else {
      # page
      &attic($config->{data}, "$page", "${page},v");
      unlink &untaint("$config->{data}/$page") or die "Could not remove page $page: $!\n";
      unlink &untaint("$config->{data}/${page},v") or die "Could not remove page $page: $!\n";
      &pageindex;
      print qq(<p class="nolink">Removed successfully page <b>$page</b>.</p>);
    }
  }
}

sub restore_attic_admin {
  return if $config->{read_only};
  my $got = 0;
  print qq(
     <h3>Restore removed files from .attic/ directories</h3>
     <form title="attic" method="post" action="$me">
        <input type="hidden" name="state" value="do_restore_attic_admin">
  );

  if (-d "$config->{data}/.attic") {
    print qq(<h4>Removed PodWiki Pages</h4><select name="pages" size="10" multiple>);
    opendir D, "$config->{data}/.attic" or die "Could not read DATA/.attic directory $config->{data}/.attic: $!\n";
    foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
      $got = 1;
      next if $page =~ /,v$/;
      my $rcs = new PodWiki::Revision(@rcs, ".attic/$page");
      my $revision = $rcs->head;
      my $revdate = scalar localtime($rcs->revdate($revision));
      print qq(<option value="$page">$page (revision $revision, $revdate)</option>);
    }
    closedir D;
    print qq( </select><br />);
  }

  if (-d "$config->{gfx}/.attic") {
    print qq(<h4>Removed Files</h4><select name="files" size="10" multiple>);
    opendir D, "$config->{gfx}/.attic" or die "Could not read GFX/.attic directory $config->{gfx}/.attic: $!\n";
    foreach my $file (sort grep { $_ !~ /^\./ } readdir(D)) {
      $got = 1;
      print qq(<option>$file</option>);
    }
    closedir D;
    print qq( </select><br />);
  }

  if ($got) {
    print qq(<br /><input type="submit" name="submit" value="Restore selected files"></form>);
  }
  else {
    print qq(<p>nothing to restore</p>);
  }
}

sub do_restore_attic_admin {
  return if $config->{read_only};
  my @pages = $query->param("pages");
  my @files = $query->param("files");

  foreach my $page (@pages) {
    system ("cp '$config->{data}/.attic/$page' '$config->{data}/'") and
      die "Unable to restore $page: $!\n";
    system ("cp '$config->{data}/.attic/${page},v' '$config->{data}/'") and
      die "Unable to restore ${page},v: $!\n";
    unlink "$config->{data}/.attic/$page" or die "Could not remove .attic/$page: $!\n";
    unlink "$config->{data}/.attic/${page},v" or die "Could not remove .attic/${page},v: $!\n";
    print qq(<p class="nolink">Restored page $page</p>);
  }

  foreach my $file (@files) {
    system ("cp '$config->{gfx}/.attic/$file' '$config->{gfx}/'") and
      die "Unable to restore $file: $!\n";
    unlink "$config->{gfx}/.attic/$file" or die "Could not remove .attic/$file: $!\n";
    print qq(<p class="nolink">Restored file $file</p>);
  }
}

sub search {
  print qq(
   <h3>Page Search</h3>
    <form title="search" action="$me" method="post">
     <input type="hidden" name="state" value="dosearch">
     <input type="text" name="string" size="30">
     <input type="submit" name="submit-full" value="full text search">
  );

  if (-e "$config->{data}/.podindex") {
    print qq(<input type="submit" name="submit-index" value="index search">);
  }

  my $a_size = scalar keys %{$attributes};
  if ($a_size) {
    # attributes exists, so provide attribute search
    print "<p>&nbsp;</p><h3>Attribute Search</h3>";
    print qq(<table border="0"><tr valign="top">);
    my $sel;
    foreach my $attribute (sort keys %{$attributes}) {
      next if $attribute =~ /^\./;
      print qq(<td><b>$attribute</b><br />);
      print qq(<select name="podattributes-${attribute}" size="10"><option></option>);
      foreach my $attr_value (sort split /\s\s*/, $attributes->{$attribute}) {
	print qq(<option>$attr_value</option>);
      }
      print qq(</select></td>);
    }
    print qq(</tr></table>);
    print qq(<input type="submit" name="submit-attributes" value="attribute search">);
  }

  print qq(</form>);
}

sub dosearch {
  if (exists $input->{"submit-full"}) {
    print qq(<h3>full text search results</h3>);
    opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
    my @res;
    my $tree = new Pod::Tree;
    print "<dl>\n";
    foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
      next if($page =~ /,v$/);
      if (-T "$config->{data}/$page") {
	open F, "<$config->{data}/$page" or die "Could not read page $page: $!\n";
	my $rcs = new PodWiki::Revision(@rcs, "$page");
	my $revision = $rcs->head;
	my $modified = scalar localtime($rcs->revdate);
	my $author   = $rcs->author;
	my $content  = join "", <F>;
	close F;
	$content =~ s/^=[a-z0-9]+//mg;
	$tree->load_string($content);
	my $plain = $tree->get_root->get_deep_text;
	if ($plain =~ /$input->{string}/i) {
	    print qq(<dt>);
	    print PodWiki::Runner->pod_link("$page");
	    print qq(
                     <font color="#c0c0c0" size="1"><b>$revision</b> by $author ($modified)</font>
                   </dt>
          );
	  my ($front, $match, $back) = $plain =~ /(.{0,30})($input->{string})(.{0,30})/i;
	  print "<dd>$front<b>$match<\/b>$back<p>&nbsp;</p><\/dd>\n";
	}
      }
    }
    print "</dl>\n";
    closedir D;
  }
  elsif (exists $input->{"submit-index"}) {
    print qq(<h3>index search results</h3>);
    &tie_index;
    my %match;
    foreach my $word (split /\s\s*/, lc($input->{string})) {
      if (exists $index{$word}) {
	map { $match{$_}++ } keys %{$index{$word}};
      }
    }
    foreach my $matchpage (sort { $match{$b} <=> $match{$a} } keys %match) {
      print "<p>" . PodWiki::Runner->pod_link($matchpage) . "</p>\n";
    }
  }
  elsif (exists $input->{"submit-attributes"}) {
    print qq(<h3>attribute search results</h3>);
    my %match;
    foreach my $attribute (sort keys %{$attributes}) {
      next if $attribute =~ /^\./;
      if ($input->{"podattributes-${attribute}"}) {
	if (exists $attributes->{".classified"}->{$attribute}->{ $input->{"podattributes-${attribute}"} }) {
	  map { $match{$_}++ } keys %{$attributes->{".classified"}->{$attribute}->{ $input->{"podattributes-${attribute}"} }};
	}
      }
    }
    foreach my $matchpage (sort { $match{$b} <=> $match{$a} } keys %match) {
      print "<p>" . PodWiki::Runner->pod_link($matchpage) . "</p>\n";
    }
  }
}

sub upload {
  return if $config->{read_only};
  my $fh = $input->{file};
  open F, "> $config->{gfx}/$input->{file}" or die "Could not write file $config->{gfx}/$input->{file}: $!\n";
  binmode F;
  my $buffer;
  while (read($fh,$buffer,1024)) {
    print F $buffer;
  }
  close F;
  &edit;
  print "<p>File $input->{file} successfully uploaded.</p>\n";
}

sub checkout {
  return if $config->{read_only};
  # check out a revision of a page
  my $rcs = new PodWiki::Revision(@rcs, "$page");
  unlink &untaint("$config->{data}/$page");

  $rcs->checkout("$input->{revision}");
  $rcs->rcs("-l");
  $rcs->commit("RESTORED from revision $input->{revision}");
  $rcs->UNLOCK();
  &display;

  print "<p>Revision $input->{revision} of page <b>$page</b> successfully restored.</p>\n";
}

sub diff {
  # print a diff between two revisions of a page
  print "<p><div class=pdotted>";
  my $rcs = new PodWiki::Revision(@rcs, "$page");
  if (! $rcs->rcsdiff("-r$input->{previous}", "-r$input->{current}")) {
    print "<p>no changes between revision $input->{previous} and $input->{current}</p>";
  }
  else {
    my (undef, undef, @diff) = $rcs->rcsdiff("-r$input->{previous}", "-r$input->{current}", "-u");
    print qq(<table border="0" width="100%" cellspacing="0" cellpadding="0">);
    local $_;
    my $color;
    my $prefix;
    foreach (@diff) {
      next if /^\@\@/;
      if (/^\+/) {
	$color = "bgcolor=\"lightgreen\"";
	$prefix = "+";
      }
      elsif (/^\-/) {
	$color = "bgcolor=\"orange\"";
	$prefix = "-";
      }
      else {
	$color = "bgcolor=\"#cccccc\"";
	$prefix = "&nbsp;";
      }
      s/\</&lt;/g; # FIXME:     should call general fixup method
      s/^[\+\-]/ /;
      print qq(<tr><td $color width="1">$prefix</td><td bgcolor="#cccccc"><pre>&nbsp;$_</pre></td></tr>);
    }
    print qq(</table>);
  }
  print "</div>\n";
  print qq(<table border="0" width="10%">
    <tr><td bgcolor="lightgreen">added</td></tr>
    <tr><td bgcolor="orange">removed</td></tr>
    <tr><td bgcolor="#cccccc">unchanged</td></tr>
    </table>
  );

  &info;
}

sub view {
  # check out a temporary revision of a page and display it
  my $dir = "/tmp/.podwiki";
  system("mkdir -p $dir") and die "Could not create temporary directory $dir: $!\n";

  my $file = &untaint("$config->{data}/${page},v");
  system("cp '$file' $dir/");

  $file = &untaint("$dir/${page},v");
  system("chmod 644 '$file'");
  my $rcs = new PodWiki::Revision(
				  bindir  => "$config->{rcsbindir}",
				  workdir => "$dir",
				  debug   => 0,
				  file    => "$page");
  $rcs->checkout("$input->{revision}");
  open F, "$dir/$page" or die "Could not read temporary revsion $input->{revision} of $page: $!\n";
  my $content = join "", <F>;
  close F;

  unlink &untaint("$dir/$page");
  unlink &untaint("$dir/${page},v");

  #### $content =~ s/\r//g;
  print qq(
	   <h3>View Revision $input->{revision} of page);
  print PodWiki::Runner->pod_link($page) . ":\n";
  print qq(
    </h3><div class="pdotted">);
  $output = "";
  &render($content);
  print $output;
  print "</div><p>&nbsp;</p>";

  &info;
}

sub register {
  return if $config->{read_only};
  print qq(
   <h3>Register New User</h3>
   <form title="register" action="$me" method="post">
   <input type="hidden" name="state" value="do_register">
    <table border="0">
      <tr>
        <th align="right">User</th>
	<td><input type="text" size="20" name="user"></td>
      </tr>
      <tr>
        <th align="right">Email</th>
	<td><input type="text" size="20" name="mail"></td>
      </tr>
      <tr>
        <th align="right">Password</th>
	<td><input type="password" size="20" name="passwd1"></td>
      </tr>
      <tr>
        <th align="right">Repeat Password</th>
	<td><input type="password" size="20" name="passwd2"></td>
      </tr>

      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Register"></td>
      </tr>
    </table>
   </form>
  );
}

sub do_register {
  return if $config->{read_only};
  if (! $input->{user}) {
    print qq(<p class="nolink">Please supply a username!</p>);
    &register;
  }
  else {
    if ($input->{passwd1} ne $input->{passwd2}) {
      print qq(<p class="nolink">Supplied passwords are not identical!</p>);
      &register;
    }
    else {
      # try to create the user
      &tie_users;
      if (exists $users->{$input->{user}}) {
	print qq(<p class="nolink">User <b>$input->{user} already exists. Please try another username.</p>);
	&register;
	return;
      }
      if ($input->{user} =~ /[`|\/\\\.\'\"]/) {
	# The characters at the end of this line are to get the emacs syntax highlighting back in sync: `]
	print qq(<p class="nolink">Username contained forbidden characters. Please try another username.</p>);
	&register;
	return;
      }
      # ok
      &save_user($input->{user}, $input->{passwd1}, $input->{mail});
      print qq(<p class="nolink">New User <b>$input->{user}</b> registered</p>);

      &render(&fetch_page("UserHome_$input->{user}"));
      return;
    }
  }
}

sub login {
  # resent the login screen and preserve CGI variables
  $input->{nextmessage} ||= "Login to PodWiki";
  $input->{nextstate}   ||= "";

  if (! $config->{read_only}) {
    print qq(
     <h3>$input->{nextmessage}
	   <a class="dotted" href="podwiki.pl?state=register">Register Yourself</a>);
    print PodWiki::Runner->pod_link("List Users|UserHomePages");
  }

  print qq(
   </h3>
   <form title="login" action="$me" method="post">
   <input type="hidden" name="state" value="check_login">
   <input type="hidden" name="nextstate" value="$input->{nextstate}">
   <input type="hidden" name="page" value="$page">
    <table border="0">
      <tr>
        <th align="right">Login</th>
	<td><input type="text" size="23" name="user"></td>
      </tr>
      <tr>
        <th align="right">Password</th>
	<td><input type="password" size="23" name="passwd"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
	<td>
           <input style="width: 150;" type="submit" name="submit" value="Login to PodWiki">
        </td>
      </tr>
    </table>
   </form>
  );
}

sub logout {
  &set_cookie($current_user, "X", "-d1");
}

sub change_passwd {
  my ($user, $message) = @_;
  if (!$user) {
    $user = $current_user;
  }
  if (! $message) {
    $message = "Change password";
  }
  if (! $user) {
    # no user to change at all
    print qq(<p class="nolink">Please <a href="$me?state=login">login</a> first.</p>);
    return;
  }
  my(undef, $pw, $mail) = &fetch_user($user);
  print qq(
   <h3>$message</h3>
   <form title="change_passwd" action="$me" method="post">
   <input type="hidden" name="state" value="do_change_passwd">
    <table border="0">
      <tr>
        <th align="right">User</th>
	<td><input type="text" size="20" name="changeuser" value="$user" readonly></td>
      </tr>
      <tr>
        <th align="right">Email</th>
	<td><input type="text" size="20" name="mail" value="$mail"></td>
      </tr>
      <tr>
        <th align="right">Old password</th>
	<td><input type="password" size="20" name="passwd_old"></td>
      </tr>
      <tr>
        <th align="right">New password</th>
	<td><input type="password" size="20" name="passwd_new"></td>
      </tr>
      <tr>
        <th align="right">New password - repeat</th>
	<td><input type="password" size="20" name="passwd_repeat"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save Changes">
        <p class="pdotted">If you want to keep the current password, do not fill out the
        password fields!</p>
        </td>
      </tr>
    </table>
   </form>
  );

  # provide preferences form
  print qq(
    <p>&nbsp;</p>
    <h3>User Preferences</h3>
    <form title="saveprefs" action="$me" method="post">
      <input type="hidden" name="state" value="saveprefs">
      <input type="hidden" name="prefuser" value="$user">
     <table border="0">
      <tr>
        <th align="right">Entry Point</th>
	<td>
         <select name="entry">
  );

  opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
  my $selected = "";
  foreach my $page (sort grep { $_ =~ /^Auto/ } readdir(D)) {
    next if $page =~ /,v$/;
    if ($page eq $entry) {
      $selected = "selected";
    }
    print qq(<option $selected value="$page">$page</option>);
    $selected = "";
  }

  closedir D;

  print qq(</select></td>
      </tr>
      <tr>
        <th align="right">Default Editor Box Height</th>
	<td><input type="text" size="20" name="height" value="$height"></td>
      </tr>
      <tr>
        <th align="right">Default Editor Box Width</th>
	<td><input type="text" size="20" name="width" value="$width"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save Preferences"></td>
      </tr>
    </table>
    </form>
  );
}


### POD parser ###

sub render {
  my $input = shift;
  my $break_mode = "break";
  my $tree = new Pod::Tree;

  next if ($input =~ /^\s*$/s); # do not render empty pages

  # very first thing to do:
  # check if it is POD, otherwise
  # use the default non-POD formatter.
  if ($input !~ /^=($POD)/m) {
    $input = "=begin $config->{default_nonpod_formatter}\n\n" . $input . "\n\n=end\n";
  }

  $input = &interpolate($input);
  $tree->load_string($input);
  $tree->walk(\&walker);
}


sub walker {
  my $node = shift;
  my $type = $node->get_type;
  my $sub = "walk_" . $type;
  &$sub($node);
}


sub walk_root {
  my $node = shift;
  &walk_children($node);
}


sub walk_children {
  my $node     = shift;
  my $children = $node->get_children;

  foreach my $child (@$children) {
      &walker($child);
  }
}


sub walk_verbatim {
  my $node = shift;
  my $text = $node->get_text;
  print &tag_open($syntax{verbatim});

  $text =~ s/\s\s*$/ /gs;  # remove trailing spaces
  $text =~ s/</&lt;/gs;    # replace <

  print $text;
  print &tag_close($syntax{verbatim});
  print "\n";
}


sub walk_ordinary {
  my $node = shift;
  my $text = $node->get_raw;
  if ($text !~ /^\s*P<[^>]*>/s && $text !~ /P<[^>]*>\s*$/s) {
    # normal paragraph
    print &tag_open($syntax{text});
    &walk_children($node);
    print &tag_close($syntax{text});
  }
  else {
    # include only page
    &walk_children($node);
  }
  print "\n";
}


sub walk_command {
  my $node    = shift;
  my $command = lc($node->get_command);
  if ($command =~ /(head[1-4])/) {
    my $target = &clean_head($node->get_deep_text);
    print "<a  name=\"$target\"></a>";
    print &tag_open($syntax{$command});
    &walk_children($node);
    print &tag_close($syntax{$command});
  }
  elsif ($command =~ /over/) {
    &walk_list($node);
  }
}

sub walk_letter {
  # workaround for node-type 'letter', which
  # is handled in walk_sequence()
  &walk_sequence(@_);
}

sub walk_sequence {
  my $node   = shift;
  my $letter = lc($node->get_letter);

  if ($letter =~ /i|b|c|f/) {
    # format element
    print &tag_open($syntax{$letter});
    &walk_children($node);
    print &tag_close($syntax{$letter});
  }
  elsif ($letter =~ /x/) {
    # index element
    if ($saving) {
      foreach my $word (split /\s\s*/, lc($node->get_deep_text)) {
	if (exists $index{$word}) {
	  # add current page to this word
	  my $H = $index{$word};
	  $H->{$page} = 1;
	  $index{$word} = $H;
	}
	else {
	  # add the word (along with current page)
	  $index{$word} = { $page => 1 };
	}
      }
    }
  }
  elsif ($letter =~ /p/) {
    # include another pod page here
    my $targetpage = $node->get_deep_text;
    if (-e "$config->{data}/$targetpage") {
      &render(&fetch_page($targetpage));
    }
    elsif ($targetpage =~ /^(intern|state):(.*)$/) {
      my ($target, $cgi_params) = split/\?/, $2;
      if ($target eq "state") {
	if ($sub2go) {
	  # internal state (== sub)
	  my $sub = "sub2go:$sub2go";
	  if (! exists $seen{$sub}) {
	    $seen{$sub} = 1;
	    &$sub2go;
	  }
	}
	else {
	  # no state, means display a page
	  &render(&fetch_page($page));
	}
      }
      else {
	if ($cgi_params) {
	  my @pairs = split /&/, $cgi_params;
	  foreach my $pair (@pairs) {
	    my ($opt, $val) = split /=/, $pair;
	    $input->{$opt} = $val;
	  }
	}
	&$target; # just run the internal sub, i.e. display or pageinfo
      }
    }
    elsif ($targetpage =~ /^http:\/\//i) {
      # fetch external content
      print &fetch_http($targetpage);
    }
    else {
      # new, non-existing page, present an "add this page" link
      my $class;
      if (-e "$config->{data}/.attic/$targetpage") {
	$class = qq(style="background: red;");
      }
      else {
	$class = qq(class="nolink");
      }
      print qq(P&lt;<font $class>$targetpage</font><a href="$me?page=$targetpage&state=edit&inherit=$page">?</a>>);
    }
  }
  elsif ($letter =~ /g/) {
    # graphic element
    my ($image, $alt) = split/\|/, $node->get_deep_text;
    if ($image !~ /\//) {
      # podwiki-maintained image in data/ directory
      $image = "$config->{gfx}/$image";
    }
    $alt ||= "$image";
    print "<img src=\"$image\" border=\"0\" alt=\"$alt\" />";
  }
  elsif ($letter =~ /t/) {
    # thumbnailing graphic element
    my ($image, $alt) = split/\|/, $node->get_deep_text;
    $alt ||= "$image";
    print qq(<a href="$config->{gfx}/$image"><img 
        src="$me?state=imagethumbnail&image=$image&size=60" 
        border="1" style="border-style:dotted" alt="$alt" /></a>);
  }
  elsif ($letter =~ /s/) {
    # non-breakable html output
    $break_mode = "nobreak";
    &walk_children($node);
    $break_mode = "break";
  }
  elsif ($letter =~ /l/) {
    # link
    my($uri, $name);
    $uri = $name = "";
    my $target       = $node->get_target;
    my $linkpage     = $target->get_page;
    ($uri, $name) = split /\s\s*/, $linkpage, 2;
    my $section      = $target->get_section;

    if ($uri =~ /^(http|https|ftp):\/\/(.*)/) {
      # external uri
      my $ziel = $2;
      my $marker = $external_link_marker;
      if ($ziel =~ /^[\/\.]/) {
	# local path, remove http:// part
	$uri = $ziel;
	$marker = $internal_link_marker; # Change to internal link marker
      }
      print "${marker}<a href=\"$uri";
    }
    elsif ($uri =~ /^mailto:(.*)$/) {
      # mail uri
      # FIXME: fragment addresse here?!
      my $address = $1;
      print "${external_link_marker}<a href=\"mailto:$address";
    }
    elsif ($uri =~ /^(state|intern):(.*)$/) {
      # internal state uri
      my ($state, $cgi_params) = split/\?/, $2;
      print "${internal_link_marker}<a href=\"$me?state=" . lc($state);
      if ($cgi_params) {
	print "&$cgi_params";
      }
    }
    else {
      # internal uri (podwiki word)
      if ($uri =~ m/(.*?)\#(.*)/) {
	$uri = $1;
	$section = $2;
      }
      if (-e "$config->{data}/$linkpage") {
	# linkpage exists
	print "${internal_link_marker}<a href=\"";
	if (exists $input->{linkhtml}) {
	  $uri = "${uri}.html";
	}
	if ($config->{direct}) {
	  print "$uri";
	}
	else {
	  print "$me?page=$uri";
	}
      }
      else {
	# nope
	$name = "?";
	my $class;
	if (-e "$config->{data}/.attic/$uri") {
	  $class = qq(style="background: red;");
	}
	else {
	  $class = qq(class="nolink");
	}
	print qq(<font $class>$uri</font><a href="$me?page=$uri&state=edit&inherit=$page);
      }
    }
    if ($section) {
      # anchor reference
      print "#" . $section;
    }
    print "\">";
    if ($name) {
      print $name;
    }
    else {
      &walk_children($node);
    }
    print "</a>";
  }
  elsif ($letter =~ /e/) {
    # character entity, html unicode and pod entities are supported
    my $entity = $node->get_deep_text;
    print "&" .$entity . ";";
  }
  elsif ($letter =~ /d/) {
    # form field
    my $name = $node->get_deep_text;
    print qq(<input type="text" name="$name" size="50">\n);
  }
  elsif ($letter =~ /m/) {
    # form field
    my $name = $node->get_deep_text;
    print qq(<textarea name="$name" rows="5" cols="60"></textarea>\n);
  }
}


sub walk_text {
  my $node  = shift;
  my $text  = $node->get_text;
  if ($break_mode eq "nobreak") {
    $text =~ s/\s/\&nbsp;/g;
  }

  print $text;
}


sub table_opt {
  my $opt = shift;
  if (! $opt) {
    return "";
  }
  else {
    $opt =~ s/^\s*//gs;
    $opt =~ s/\s*$//gs;
    $opt =~ s/,/ /g;
    $opt = " $opt";
    return $opt;
  }
}


sub walk_form {
  my $node = shift;
  print qq(
    <form title="PodWiki" action="$me" method="post">
     <h3>Create Page from Template <img src="$config->{gfx}/ilink.png" alt="Internal Link" />
     <a href="$me?page=$page">$page</a>: <input type="text" name="setpage"></h3>
     <input type="hidden" name="state" value="save">
     <input type="hidden" name="template" value="$page">
     <p>Get some help about
        <img src="$config->{gfx}/ilink.png" alt="Internal Link" /> <a href="$me?page=PodWikiMarkup">PodWikiMarkup</a>
        or <img src="$config->{gfx}/ilink.png" alt="Internal Link" /> <a href="$me?page=WikiShorthand">WikiShorthand</a></p>
     <div class="pdotted">
  );

  &walk_children($node);

  # FIXME: this is much identical code as in edit() - outsourcing into extra sub?
  print qq(
    </div>
    Comment:<br />
    <input type="text" name="comment" class="dotted" size="70"><br /><br />
    State: <select name="rcs_state">
      <option selected>Experimental</option>
      <option>Testing</option>
      <option>Bugfix</option>
      <option>Final</option>
    </select>
  );

  my $a_size = scalar keys %{$attributes};
  my $attribute_source = $page; # use form template as the attribute source
  if ($a_size) {
    # attributes exists, so provide them
    print "<h4>Page Classification:</h4>";
    print qq(<table frame="box" rules="group" width="50%" border="1">);
    my $sel;
    foreach my $attribute (sort keys %{$attributes}) {
      next if $attribute =~ /^\./;
      foreach my $attr_value (sort split /\s\s*/, $attributes->{$attribute}) {
        if (exists $attributes->{".classified"}->{$attribute}->{$attr_value}->{$attribute_source}) {
          print qq(
             <tr valign="top">
               <th>$attribute:</th>
               <td>$attr_value</td>
               <input type=hidden name="podattributes-${attribute}" value="$attr_value">
             </tr>
          );
        }
      }
    }
    print qq(</table><br />);
  }

  print qq(
 <input type="submit" name="save" value=" &nbsp; Save &nbsp; ">
 <a href="$me?page=$page">Cancel</a>
 </form>
 );
}

sub walk_field {
  my $node = shift;
  my ($type, $options) = split/ /, &table_opt($node->get_arg), 2;
  &walk_children($node);
  if ($type =~ /type=textarea/) {
    print qq(<textarea $options></textarea>\n);
  }
  else {
    print qq(<input $options>\n);
  }
}

sub walk_table {
  my $node = shift;
  my $options = &table_opt($node->get_arg);
  print &tag_open($syntax{table} . $options) . "\n";
  &walk_children($node);
  print &tag_close($syntax{table}) . "\n";
}

sub walk_row {
  my $node = shift;
  my $options = &table_opt($node->get_arg);
  print &tag_open($syntax{row} . $options) . "\n";
  &walk_children($node);
  print &tag_close($syntax{row}) . "\n";
  &walk_siblings($node);
}

sub walk_cell {
  my $node = shift;
  my $tag = "cell";
  my $options = &table_opt($node->get_arg);
  if ($options) {
    # removed + match
    $options =~ s/type=head//;
    $tag = "headcell";
  }
  print &tag_open($syntax{$tag} . $options) . "\n";
  print $node->get_deep_text;
  &walk_siblings($node);
  print &tag_close($syntax{$tag}) . "\n";
}

sub walk_list {
  my $node      = shift;
  my $indent    = $node->get_arg;
  my $list_type = $node->get_list_type;

  print &tag_open($syntax{"list_" . $list_type});

  &walk_children($node);	# text of the =over paragraph

  print &tag_close($syntax{"list_" . $list_type});
}


sub walk_item {
  my $node      = shift;
  my $item_type = $node->get_item_type;

  if ($item_type eq "text") {
    # indented text, handle different than ordinary lists
    print "<dt>";
    &walk_children($node);	# text of the =item paragraph
    print "</dt>";
    print &tag_open($syntax{"item_" . $item_type});
    &walk_siblings($node);
    print &tag_close($syntax{"item_" . $item_type});
  }
  else {
    # ordinary list
    print &tag_open($syntax{"item_" . $item_type});
    if ($item_type ne "bullet" && $item_type ne "number") {
      # bullet types do not have content on the =item line beside the bullet
      &walk_children($node);	# text of the =item paragraph
      print "<br />";
    }
    &walk_siblings($node);
    print &tag_close($syntax{"item_" . $item_type});
  }
}


sub walk_for {
  my $node      = shift;
  my $formatter = $node->get_arg;
  my $text      = $node->get_text;
  $formatter =~ s/\s//g;

  # call the generalized formatter
  $fmttr->format_as($formatter, $text);
}

sub walk_siblings {
  my $node     = shift;
  my $siblings = $node->get_siblings;

  for my $sibling (@$siblings) {
    &walker($sibling);
  }
}


sub tag_open {
  my $name = shift;
  return "<" . $name . ">";
}


sub tag_close {
  my $name = shift;
  return "</" . $name . ">";
}


sub data_empty {
  opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
  my @entries = grep { $_ !~ /^\./ } readdir(D);
  closedir D;
  my $num = scalar @entries;
  if ($num >= 1) {
    return 0; # not empty
  }
  else {
    return 1; # empty
  }
}

sub setup {
  return if $config->{read_only};
  # initially set up things

  # do not execute setup if already done so
  return if(! &data_empty);

  # create the content
  opendir D, "$config->{templates}" or die "Could not read template directory: $!\n";
  foreach my $file (grep { $_ !~ /^\./ } readdir(D)) {
    system("cp", &untaint("$config->{templates}/$file"), &untaint("$config->{data}/$file")) and
      die "Could not install template file $file: $!\n";
    my $rcs = new PodWiki::Revision(@rcs, "$file");
    $rcs->add('InitialCheckinByPodWikiSetup');
    $rcs->checkout();
  }

  &tie_users;
  &tie_groups;
  &save_user($config->{default_admin_user}, $config->{default_admin_passwd});
  &set_member($config->{default_admin_user}, $config->{default_admin_group});
}



sub tie_variables {
  my $ref = PodWiki::Struct::fetch("$config->{data}/.podvariables");
  %variables = %{ $ref };
}


sub tie_attributes {
  $attributes = PodWiki::Struct::fetch("$config->{data}/.podattributes");
}

sub tie_index {
  my $ref = PodWiki::Struct::fetch("$config->{data}/.podindex");
  %index = %{ $ref };
}

sub tie_comment {
  my $ref = PodWiki::Struct::fetch("$config->{data}/.podcomments");
  %comment = %{ $ref };
}


sub tie_users {
  $users = PodWiki::Struct::fetch("$config->{data}/.podusers");
}

sub tie_groups {
  $groups = PodWiki::Struct::fetch("$config->{data}/.podgroups");
}

sub store_attribute {
  return if $config->{read_only};
  my($attribute, $value) = @_;

  foreach my $attr_value (keys %{$attributes->{".classified"}->{$attribute}}) {
    # remove from other values, if any
    delete $attributes->{".classified"}->{$attribute}->{$attr_value}->{$page};
  }

  if ($value) {
    # only add it if something has been selected
    $attributes->{".classified"}->{$attribute}->{$value}->{$page} = 1;
  }

  PodWiki::Struct::store($attributes, "$config->{data}/.podattributes");
}


sub parse_options {
  # parse config options per page
  my $page = shift;
  if (-e "$config->{data}/$page") {
    if (! $page_conf_done) {
      # do it only if not already done so
      open P, "$config->{data}/$page" or die "Could not open page $page for reading: $!\n";
      $page_conf = {};
      local $_;
      my $o;
      while (<P>) {
	next if(/^\s*$/);
	next if(/^\s*#/);
	chomp;
	if (/^=(begin|for) options/) {
	  $o = 1;
	}
	elsif ($o) {
	  if (/^=(head|end|cur|over|begin)/) {
	    last;
	  }
	  else {
	    s/^\s*//g;           # remove leading whitespace
	    s/\s*$//g;           # remove trailing whitespace
	    s/\s*(?<!\\)#.*$//;  # remove trailing comment, if '#' is not prepended(masked) with \
	    my($opt, $val) = split /\s*=\s*/, $_, 2;
	    $page_conf->{$opt} = $val;
	  }
	}
      }
      close P;
      $page_conf_done = 1;
    }
  }
  $page_conf;
}

sub fetch_http {
  my $uri = shift;
  if ($uri =~ /^http:\/\/(.*)$/) {
    my ($host, $path) = split /\//, $1, 2;
    my (undef, $port) = split /:/, $host, 2;
    $path = "/" . $path;
    $port ||= 80;
    my $conn = new IO::Socket::INET(
                                  Proto    => "tcp",
                                  PeerAddr => $host,
                                  PeerPort => $port,
                                 ) or die "Unable to connect to $host:$port: $!\n";
    $conn->autoflush(1);

    print $conn qq(GET $path HTTP/1.0
User-Agent: Mozilla/5.0 (compatible; PodWiki::fetch_http $PodWiki::VERSION)
Host: $host
Accept: text/*;q=1.0, image/png;q=1.0, image/jpeg;q=1.0, image/gif;q=1.0, image/*;q=0.8, */*;q=0.5
Accept-Charset: iso-8859-1;q=1.0, *;q=0.9, utf-8;q=0.8


);
    my $content = join "", <$conn>;
    close $conn;

    return  "<!-- START - external content included by intern:fetch_http. uri: $uri-->\n"
          . &clean_http_content($content)
	  . "<!--   END - external content included by intern:fetch_http. uri: $uri-->\n";
  }
}


sub clean_http_content {
  # FIXME: replace <img src=..> uri's with remote hostname
  my $content = shift;

  # remove header fields
  $content =~ s/\r//g;
  $content =~ s/^.+?\n\n//s;

  # remove redundant HTML tags
  $content =~ s/(<html>|<body[^>]*>|<\/html>|<\/body>)//gis;

  # remove HTML header field, if any. we are using our own
  # and this could disrupt our page design/layout/style
  $content =~ s/<head>.*<\/head>//is;

  return $content;
}


sub imagethumbnail {
  print "Content-type: image/png\n";
  binmode STDOUT;
  my $img    = &untaint("$config->{gfx}/$input->{image}");
  my $identify = &untaint("$config->{image_magick}/identify");
  my $convert  = &untaint("$config->{image_magick}/convert");
  my (undef, undef, $size, @undef) = split / /, `$identify $img`;
  $size =~ s/\+.*$//;
  my($x, $y) = split /x/, $size;

  if (exists $input->{size}) {
    $size = $input->{size};
  }
  else {
      if (defined($config->{thumbnailsize})) {
	  $size = $config->{thumbnailsize};
      } else {
	  # Continue to accept the "tumbsize" name
	  $size = $config->{tumbsize};
      }
  }

  if ($x > $size && $y > $size) {
    my ($image, $create);
    $create = 1;
    if (-e "$config->{gfx}/$input->{image}-thumbnail") {
      # cached thumbnail exists
      my $mt_thumb = (stat("$config->{gfx}/$input->{image}-thumbnail"))[9];
      my $mt_image = (stat("$config->{gfx}/$input->{image}"))[9];
      if ($mt_thumb > $mt_image) {
	# image unchanged, use the cached version
	open TH, "< $config->{gfx}/$input->{image}-thumbnail" or
	  die "Could not open cached image thumbnail: $config->{gfx}/$input->{image}-thumbnail: $!\n";
	$image = join "", <TH>;
	close TH;
	$create = 0;
      }
      else {
	$create = 1;
      }
    }

    if ($create) {
      my $size = &untaint("${size}x${size}");
      $image = `$convert -resize ${size}x${size} $img png:-`;
      my $th = &untaint("$config->{gfx}/$input->{image}-thumbnail");
      open TH, "> $th" or
	  die "Could not open cached image thumbnail: $config->{gfx}/$input->{image}-thumbnail: $!\n";
      binmode TH;
      print TH $image;
      close TH;
    }

    print "Content-length: " . length($image) . "\n\n";
    print $image;
  }
  else {
    my $image = `$convert $img png:-`;
    print "Content-length: " . length($image) . "\n\n";
    print $image;
  }
}

sub formatters {
  # print a list of all installed formatters 
  print "The following formatter tags are recognized on this PodWiki:\n";
  print "<ul>\n";
  foreach my $fmttr ($fmttr->tags()) {
    print "<li>$fmttr</li>\n";
  }
  print "</ul>\n";
}

sub dumper {
  print "<pre>" . Dumper(shift()) . "</pre>\n";
}

sub detach {
  # save the dbm file (if any changes were made)
  # so that subsequent calls of Runner deliver always
  # the current content and no cached version of the
  # dbm hashes (under mod_perl!)
}

sub attic {
  my($path, @files) = @_;
  if (! -d "$path/.attic") {
    my $dir = &untaint("$path/.attic");
    system("mkdir", "$dir") and die "Unable to 'mkdir $path/.attic': $!\n";
  }
  for my $file (@files) {
    my $src = &untaint("$path/$file");
    my $dst = &untaint("$path/.attic/$file");
    system ("cp", "$src", "$dst") and die "Unable to copy $file to .attic/: $!\n";
  }
}

sub errlog {
  my ($level, $msg) = @_;
  print STDERR "[" . scalar localtime() . "] $me: " . "[$level] $msg\n";
}

sub get_lockfile {
  return &untaint("$config->{data}/.podwiki-LOCK");
}

sub lock_podwiki {
  my $lock = &get_lockfile();
  open LOCK, ">$lock" or die "Could not open LOCK file: $lock: $!\n";
  flock LOCK, LOCK_EX;
  return *LOCK;
}

sub unlock_podwiki {
  my $fd = shift;
  flock $fd, LOCK_UN;
  unlink &get_lockfile();
}

sub getrandom {
  my($len) = @_;
  my @range = ('a'..'z','A'..'Z','0'..'9');
  my $out;
  foreach (1.. $len) {
     $out .= $range[rand @range]; 
  }
  return $out;
}

sub AUTOLOAD {
  my $sub = $PodWiki::Runner::AUTOLOAD;
  $sub =~ s/.*:://;
  return if $sub eq "DESTROY";
  print qq(<p class="nolink">No such state: <b>$sub</b>!</p>);
}

sub untaint {
  if ($_[0] =~ /^(.*)$/) {
    return $1;
  }
}

END {
  &detach;
}

1; #end
