 #
 #   $Id: UI.pm,v 1.8 2005/03/24 23:02:46 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #


#
# this module provides functions which can be directly
# used by the user via P<state:WHAT> or L<state:WHAT>
#


package PodWiki::UI;

$PodWiki::UI::VERSION = "1.02";



use lib qw(. lib);

use strict;
no strict "refs";

use Exporter;
use Data::Dumper;
use CGI::Cookie;


use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS);

@ISA=qw(Exporter);
@EXPORT=qw(jump info newest pageindex fileindex toc_by_attribute toc);
@EXPORT_OK=qw();
@EXPORT_TAGS=();



sub jump {
  # print a select box to jump directly to a page
  opendir D, "$PodWiki::Runner::config->{data}" or
    die "Could not read DATA directory $PodWiki::Runner::config->{data}: $!\n";
  print qq(<form name="jump" title="jump" action="$PodWiki::Runner::me" method="get">
           <select name="page" onchange="document.forms['jump'].submit()">\n);
  foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($page =~ /\,v$/);
    next if($page =~ /\,lockuser$/);
    print qq(<option value="$page">$page</option>\n);
  }
  closedir D;
  print qq(</select><input type="submit" value="go"></form>\n);
}


sub info {
  my $rcs = new PodWiki::Revision(@PodWiki::Runner::rcs, "$PodWiki::Runner::page");
  my $revision = $rcs->head;
  my @revisions = $rcs->revisions;
  print qq(<h3>PageInfo about ${PodWiki::Runner::internal_link_marker}<i><a href="$PodWiki::Runner::me?page=$PodWiki::Runner::page">$PodWiki::Runner::page</a></i> (Revision: $revision) 
     <a class="dotted" href="$PodWiki::Runner::me?page=$PodWiki::Runner::page&state=edit">EditThisPage</a></h3>);

  my %comment = $rcs->comments;
  my $previous = 0;
  my %prev;
  foreach my $rev (reverse @revisions) {
    if ($previous) {
      $prev{$rev} = $previous;
    }
    $previous = $rev;
  }

  foreach my $rev (@revisions) {
    my $revdate = scalar localtime($rcs->revdate($rev));
    my $state   = $rcs->state($rev);
    my $author  = $rcs->author($rev);
    print qq(
     <table border="0" bgcolor="#000000" cellspacing="1" width=60%"><tr><td>
      <table border="0" cellspacing="1" cellpadding="5" width="100%">
      <tr>
        <th bgcolor="#cfcfcf" align="right" width="20%">Revision</th>
        <td bgcolor="#ffffff">
         <table border="0"><tr>
         <td>$rev</td>
         <td>
         <form title="activity" action="$PodWiki::Runner::me" method="get">
          <input type="hidden" name="page" value="$PodWiki::Runner::page">
          <input type="hidden" name="revision" value="$rev">
          <input type="hidden" name="state" value="checkout">
          <input type="submit" name="submit" value="Restore">
         </form>
         </td>
         <td>
         <a class="dotted" href="$PodWiki::Runner::me?page=$PodWiki::Runner::page&state=view&revision=$rev">View</a>
    );

    if (exists $prev{$rev}) {
      print
        qq(<a class="dotted"
            href="$PodWiki::Runner::me?page=$PodWiki::Runner::page&state=diff&previous=$prev{$rev}&current=$rev">diff to $prev{$rev}</a>
      );
    }

    print qq(
         </td></tr></table>
        </td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">Date</th><td bgcolor="#ffffff">$revdate</td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">State</th><td bgcolor="#ffffff">$state</td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">Author</th><td bgcolor="#ffffff">$author</td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">Comment</th><td bgcolor="#ffffff">$comment{$rev}</td>
      </tr>
      </table></td></tr></table><p>&nbsp;</p>
    );
    $previous = $rev;
  }
}

sub pageindex {
  # build an index of all existing pages
  opendir D, "$PodWiki::Runner::config->{data}" or die "Could not read DATA directory $PodWiki::Runner::config->{data}: $!\n";
  print qq(
   <h3>PageIndex</h3>
   <table border="0" bgcolor="#000000" cellspacing="1"><tr><td>
    <table border="0" cellspacing="1" cellpadding="5" >
     <tr bgcolor="#cfcfcf">
      <th>Page</th><th>Revision</th><th>Author</th><th>Last modified</th><th>Size</th><th>Action</th>
     </tr>
  );

  foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($page =~ /\,v$/);
    next if($page =~ /\,lockuser$/);
    my $rcs = new PodWiki::Revision(@PodWiki::Runner::rcs, "$page");
    my $revision = $rcs->head;
    my $modified = scalar localtime($rcs->revdate);
    my $author   = $rcs->author;
    my $size     = (stat("$PodWiki::Runner::config->{data}/$page"))[7];
    my $link;
    if ($PodWiki::Runner::config->{direct}) {
      $link = "$page";
    }
    else {
      $link = "$PodWiki::Runner::me?page=$page";
    }
    print qq(
      <tr bgcolor="#ffffff">
       <td><p>${PodWiki::Runner::internal_link_marker}<a href="$link">$page</a></p></td>
       <td><p>$revision</p></td>
       <td><p>$author</p></td>
       <td><p>$modified</p></td>
       <td><p>$size</p></td>
       <td>
         <p>
            <a class="dotted" href="$PodWiki::Runner::me?page=$page&state=src">source</a>&nbsp;);
    if (! $PodWiki::Runner::config->{read_only}) {
      print qq(
            <a class="dotted" href="$PodWiki::Runner::me?page=$page&state=edit">edit</a>&nbsp;
            <a class="dotted" href="$PodWiki::Runner::me?page=$page&state=remove">remove</a>);
    }
    print qq(
         </p>
       </td>
      </tr>
    );
  }
  print qq(</table></td></tr></table>
    <p>
       ${PodWiki::Runner::internal_link_marker}<a href="$PodWiki::Runner::me?state=fileindex">Maintain File Index (uploaded images/files)</a></p>
  );
  closedir D;
}


sub fileindex {
  ### images
  opendir D, "$PodWiki::Runner::config->{gfx}" or die "Could not read IMAGES directory $PodWiki::Runner::config->{gfx}: $!\n";
    print qq(
   <p>&nbsp;</p>
   <h3>FileIndex</h3>
   <table border="0" bgcolor="#000000" cellspacing="1"><tr><td>
    <table border="0" cellspacing="1" cellpadding="5" >
     <tr bgcolor="#cfcfcf">
      <th>Filename</th><th>Last modified</th><th>Edit</th>
     </tr>
  );

  my $thumbnail = "";
  foreach my $file (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($file =~ /thumbnail$/);
    my $modified = scalar localtime( (stat("$PodWiki::Runner::config->{gfx}/$file"))[9] );
    if ($PodWiki::Runner::config->{image_magick}) {
      # image magick support is enabled
      if ($file =~ /(jpg|jpeg|png|gif|bmp|xpm|tiff)$/) {  # FIXME: more suffixes to add?
	$thumbnail = qq(<img src="$PodWiki::Runner::me?state=imagethumbnail&image=$file" border="0" /><br />);
      }
    }
    print qq(
      <tr bgcolor="#ffffff">
       <td align="center"><p>${thumbnail}${file}</p></td>
       <td><p>$modified</p></td>
       <td>
        <p>
         <a class="dotted" href="$PodWiki::Runner::config->{gfx}/$file" target="__new__">view</a>&nbsp; &sdot);
    if (! $PodWiki::Runner::config->{read_only}) {
      print qq(<a class="dotted" href="$PodWiki::Runner::me?file=$file&state=remove">remove</a>);
    }
    print qq(
        </p>
       </td>
      </tr>
    );
    $thumbnail = "";
  }
  print qq(</table></td></tr></table>);

  closedir D;
}

sub newest {
  opendir D, "$PodWiki::Runner::config->{data}" or die "Could not read DATA directory $PodWiki::Runner::config->{data}: $!\n";
  print "<h3>Newest Pages</h3>\n";
  print "<ul>\n";

  my %file;
  foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($page =~ /\,v$/);
    next if($page =~ /\,lockuser$/);
    my $modified = (stat("$PodWiki::Runner::config->{data}/$page"))[9];
    $file{$page} = $modified;
  }

  my $counter = $PodWiki::Runner::input->{count} || 5;
  foreach my $page (sort { $file{$b} <=> $file{$a} } keys %file) {
    if ($counter) {
      my $modified = scalar localtime($file{$page});
      print qq(
       <li><b>$modified</b>  ${PodWiki::Runner::internal_link_marker}<a href="$PodWiki::Runner::me?page=$page">$page</a></li>
      );
    }
    else {
      last;
    }
    $counter--;
  }

  print "</ul>\n";
  closedir D;
}



sub toc {
  # generate a table of content
  my $space;
  print "<p>\n";
  foreach my $target (@{$PodWiki::Runner::head{$PodWiki::Runner::page}}) {
    $space = "&nbsp;&nbsp;&nbsp;&nbsp;" x $target->{level};
    print qq( $space<a href="#$target->{target}">$target->{title}</a><br />\n);
  }
  print "</p><p>&nbsp;</p>\n";
}

sub toc_by_attribute {
  # generate a table of content by selected attribute
  # FIXME: support for toc by multiple attributes!!!
  my ($attribute, $value);
  foreach my $attr (keys %{$PodWiki::Runner::attributes}) {
    next if $attr =~ /^\./;
    if (exists $PodWiki::Runner::input->{$attr}) {
      # this one has been set by CGI
      $attribute = $attr;
      $value     = $PodWiki::Runner::input->{$attr};
    }
  }
  print "<p>\n";
  if (exists $PodWiki::Runner::attributes->{".classified"}->{$attribute}->{$value}) {
    foreach my $page (sort keys %{$PodWiki::Runner::attributes->{".classified"}->{$attribute}->{$value}}) {
      if (-e "$PodWiki::Runner::config->{data}/$page") {
	print qq(&nbsp;&nbsp;&nbsp;&nbsp;${PodWiki::Runner::internal_link_marker}<a href="$PodWiki::Runner::me?page=$page&$attribute=$value">$page</a><br />);
      }
      else {
	my $ref = $PodWiki::Runner::attributes->{".classified"};
	delete $ref->{$attribute}->{$value}->{$page};
	$PodWiki::Runner::attributes->{".classified"} = $ref;
      }
    }
  }
  delete $PodWiki::Runner::input->{$attribute}; # clean up
  delete $PodWiki::Runner::input->{attribute};
  print "</p>\n";
}



1;
