 #
 #   $Id: Vars.pm,v 1.5 2005/03/24 23:02:46 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #
 #   Additional Copyrights:
 #
 #

package PodWiki::Vars;

use lib qw(. lib);

$PodWiki::Vars::VERSION = "1.0";

use strict;
use Carp;
use Exporter;

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS $AUTOLOAD);

@ISA=qw(Exporter);
@EXPORT=qw();
@EXPORT_OK=qw();
@EXPORT_TAGS=();


sub new {
  my($this, %param) = @_;
  my $class = ref($this) || $this;
  my $self = \%param;
  bless($self, $class);
  $self->init();
  return $self;
}


sub init {
  my($this) = @_;

  my @graphics = qw(
ascii     configure  forward    image    tar
attach    contents   encrypted  gohome   info   tgz
back      daemon     filefind   help     lock   up
bookmark  document   fileopen   html     ok     video
cancel    down       fileprint  idea     pdf
cdimage   edit       filesave   sound
		   );
  $this->{g_prep} = join "|", @graphics;


  $this->{colors} = {
		     YELLOW   => '<font color="#ffff00">',
		     ORANGE   => '<font color="#ff6600">',
		     RED      => '<font color="#ff0000">',
		     PINK     => '<font color="#ff00ff">',
		     PURPLE   => '<font color="#800080">',
		     TEAL     => '<font color="#008080">',
		     NAVY     => '<font color="#000080">',
		     BLUE     => '<font color="#0000ff">',
		     AQUA     => '<font color="#00ffff">',
		     LIME     => '<font color="#00ff00">',
		     GREEN    => '<font color="#008000">',
		     OLIVE    => '<font color="#808000">',
		     MAROON   => '<font color="#800000">',
		     BLACK    => '<font color="#000000">',
		     GRAY     => '<font color="#808080">',
		     SILVER   => '<font color="#c0c0c0">',
		     WHITE    => '<font color="#ffffff">',
		     END      => '</font>',

		     B_YELLOW   => '<span style="background-color: #ffff00">',
		     B_ORANGE   => '<span style="background-color: #ff6600">',
		     B_RED      => '<span style="background-color: #ff0000">',
		     B_PINK     => '<span style="background-color: #ff00ff">',
		     B_PURPLE   => '<span style="background-color: #800080">',
		     B_TEAL     => '<span style="background-color: #008080">',
		     B_NAVY     => '<span style="background-color: #000080">',
		     B_BLUE     => '<span style="background-color: #0000ff">',
		     B_AQUA     => '<span style="background-color: #00ffff">',
		     B_LIME     => '<span style="background-color: #00ff00">',
		     B_GREEN    => '<span style="background-color: #008000">',
		     B_OLIVE    => '<span style="background-color: #808000">',
		     B_MAROON   => '<span style="background-color: #800000">',
		     B_BLACK    => '<span style="background-color: #000000">',
		     B_GRAY     => '<span style="background-color: #808080">',
		     B_SILVER   => '<span style="background-color: #c0c0c0">',
		     B_WHITE    => '<span style="background-color: #ffffff">',
		     B_END      => '</span>'
	       };
  $this->{c_prep} = join "|", keys %{$this->{colors}};
}


sub interpolate {
  my($this, $content) = @_;
  $content =~ s/(?<!\\)(\$)($this->{g_prep})/G<${2}.png>/g;
  $content =~ s/(?<!\\)(\$)($this->{c_prep})/$this->{colors}->{$2}/g;
  return $content;
}

sub listvars {
  my($this) = @_;
  my $content = "";
  my @keys;

  @keys = split(/\|/, $this->{g_prep});
  foreach my $key (sort @keys) {
      $content .= "<tr><td> " . $key . " </td><td>" . "</td><td>Graphics</td><td>" . "G<${key}.png>" . "</td></tr>\n";
  }

  @keys = split(/\|/, $this->{c_prep});
  foreach my $key (sort @keys) {
      if ($key =~ m/B_(.*)/) {	# Background color
	  my $color = $1;
	  #lc($color);
	  if ($key eq "B_END") {
	      $content .= "<tr><td> " . $key . " </td><td>" . "Ends background color" . "</td><td>Colors</td><td>" . "" . "</td></tr>\n";
	  } else {
	      $content .= "<tr><td> " . $key . " </td><td>" . "Switches to a $color background" . "</td><td>Colors</td><td>" . "$this->{colors}->{$key} $color background $this->{colors}->{END}" . "</td></tr>\n";
	  }
      } else {	# Foreground color
	  my $color = $key;
	  #lc($color);
	  if ($key eq "END") {
	      $content .= "<tr><td> " . $key . " </td><td>" . "Ends text color" . "</td><td>Colors</td><td>" . "" . "</td></tr>\n";
	  } else {
	      $content .= "<tr><td> " . $key . " </td><td>" . "Switches to $color text" . "</td><td>Colors</td><td>" . "$this->{colors}->{$key} $color text $this->{colors}->{END}" . "</td></tr>\n";
	  }
      }
  }
  return $content;
}


1;
